from io import BytesIO
import uuid
import cavro
import pytest


from pathlib import Path


SIMPLE_CONTAINER = b'Obj\x01\x04\x14avro.codec\x08null\x16avro.schema\x0a"int"\x00aaaaaaaaaaaaaaaa\x02\x02\x02aaaaaaaaaaaaaaaa'


def test_simplest_container_invalid_magic():
    container = cavro.ContainerReader(SIMPLE_CONTAINER)
    assert list(container) == [1]


def test_container_invalid_magic():
    bad_container = b"Obj\x02" + SIMPLE_CONTAINER[4:]
    with pytest.raises(ValueError, match="Invalid file header"):
        cavro.ContainerReader(bad_container)


def test_container_invalid_codec():
    bad_container = SIMPLE_CONTAINER.replace(b'null', b'xxxx')
    with pytest.raises(cavro.CodecUnavailable, match=r"Unsupported codec: 'xxxx'"):
        cavro.ContainerReader(bad_container)


def test_container_reading_from_fileobj():
    here = Path(__file__).parent
    container_file = here / 'data' / 'weather.avro'
    container = cavro.ContainerReader(container_file.open('rb'))
    assert list(container)[2].temp == -11


class FakeUUID:

    HEADER = b'Obj\x01\x04\x16avro.schema\n"int"\x14avro.codec\x08null\x00abcdefghijklmnop'
    bytes = b'abcdefghijklmnop'


def test_writing_empty(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.close()
    assert buf.getvalue() == b'Obj\x01\x04\x16avro.schema\n"int"\x14avro.codec\x08null\x00abcdefghijklmnop\x00\x00abcdefghijklmnop'


def test_writing_empty_no_close(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    del writer
    assert buf.getvalue() == b'Obj\x01\x04\x16avro.schema\n"int"\x14avro.codec\x08null\x00abcdefghijklmnop\x00\x00abcdefghijklmnop'


def test_writing_one_int(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.write_one(1)
    writer.close()
    assert buf.getvalue() == FakeUUID.HEADER + b'\x02\x02\x02' + FakeUUID.bytes


def test_writing_one_int_no_close(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.write_one(1)
    assert buf.getvalue() == b''
    del writer
    assert buf.getvalue() == FakeUUID.HEADER + b'\x02\x02\x02' + FakeUUID.bytes


def test_writing_bigger_int(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.write_one(64)
    writer.close()
    assert buf.getvalue() == FakeUUID.HEADER + b'\x02\x04\x80\x01' + FakeUUID.bytes


def test_writing_two_ints(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.write_one(64)
    writer.write_one(1)
    writer.close()
    assert buf.getvalue() == FakeUUID.HEADER + b'\x04\x06\x80\x01\x02' + FakeUUID.bytes


def test_writing_two_blocks_of_ints(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch, max_blocksize=1)
    writer.write_one(64)
    assert buf.getvalue() == b''
    writer.write_one(1)
    assert buf.getvalue() == FakeUUID.HEADER + b'\x02\x04\x80\x01' + FakeUUID.bytes
    writer.close()
    assert buf.getvalue() == FakeUUID.HEADER + b'\x02\x04\x80\x01' + FakeUUID.bytes + b'\x02\x02\x02' + FakeUUID.bytes


def test_cannot_write_after_close():
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    writer = cavro.ContainerWriter(buf, sch)
    writer.write_one(1)
    writer.close()
    with pytest.raises(ValueError):
        writer.write_one(2)


def test_writing_two_ints_context(monkeypatch):
    monkeypatch.setattr(uuid, 'uuid4', FakeUUID)
    buf = BytesIO()
    sch = cavro.Schema('"int"')
    with cavro.ContainerWriter(buf, sch) as writer:
        writer.write_one(64)
        writer.write_one(1)
    assert buf.getvalue() == FakeUUID.HEADER + b'\x04\x06\x80\x01\x02' + FakeUUID.bytes


@pytest.mark.parametrize('codec', [c.decode() for c in cavro.CODECS.keys()])
@pytest.mark.parametrize('source_vals,schema', [
    ([1], '"int"'),
    (["The", "Cat", "Sat", "on", "the", "mat"], '"string"'),
    (["Apples"] * 10_000, '"string"'),
])
def test_round_tripping(codec, source_vals, schema):
    buf = BytesIO()
    sch = cavro.Schema(schema)
    with cavro.ContainerWriter(buf, sch, codec) as writer:
        writer.write_many(source_vals)
    buf.seek(0)
    reader = cavro.ContainerReader(buf)
    obs = list(reader)
    assert obs == source_vals



cdef class EnumType(_NamedType):
    """The avro enum type"""
    type_name = 'enum'

    cdef readonly tuple symbols
    cdef dict symbol_indexes
    cdef readonly object default_value
    cdef readonly str doc

    def __init__(self, schema, source, namespace):
        _NamedType.__init__(self, schema, source, namespace)
        cdef Schema schema_ = schema
        raw_symbols = source['symbols']
        if not isinstance(raw_symbols, (list, tuple)):
            raise ValueError("Enum symbols must be a list of strings")
        
        if schema_.options.enum_symbols_must_be_unique:
            seen_symbols = set()
            for symbol in raw_symbols:
                if symbol in seen_symbols:
                    raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
                seen_symbols.add(symbol)
        else:
            seen_symbols = raw_symbols

        name_pattern = schema_.options.name_pattern
        if schema_.options.enforce_enum_symbol_name_rules:
            for symbol in seen_symbols:
                if not isinstance(symbol, str):
                    raise InvalidName('Enum symbols must be a list of strings')
                if schema_.options.enforce_enum_symbol_name_rules:
                    if not name_pattern.fullmatch(symbol):
                        raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
        
        self.symbols = tuple(raw_symbols)
        cdef size_t i
        self.symbol_indexes = {}
        for i, symbol in enumerate(self.symbols):
            self.symbol_indexes[symbol] = i
        self.default_value = source.get('default', NO_DEFAULT)
        if self.default_value is not NO_DEFAULT:
            if self.default_value not in self.symbol_indexes:
                raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
        self.doc = source.get('doc', '')

    cpdef AvroType copy(self):
        cdef EnumType new_inst = self.clone_base()
        new_inst.symbols = self.symbols
        new_inst.symbol_indexes = self.symbol_indexes
        new_inst.default_value = self.default_value
        new_inst.doc = self.doc
        return new_inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {
            'type',
            'name', 
            'namespace', 
            'aliases', 
            'doc', 
            'symbols'
        })

    cpdef dict _get_schema_extra(self, set created):
        extra = _NamedType._get_schema_extra(self, created)
        if self.default_value is not NO_DEFAULT:
            extra['default'] = self.default_value
        extra['symbols'] = list(self.symbols)
        return extra

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef size_t index = self.symbol_indexes[value]
        zigzag_encode_long(buffer, index)

    cdef _binary_buffer_decode(self, _Reader buffer):
        return self.symbols[zigzag_decode_long(buffer)]

    cdef int _get_value_fitness(self, value) except -1:
        try:
            if value in self.symbol_indexes:
                return FIT_EXACT
        except TypeError:
            return FIT_NONE

    cpdef object _convert_value(self, object value):
        # if value is valid for type, then it needs no conversion
        return value

    cdef _json_format(self, value):
        if value not in self.symbol_indexes:
            raise KeyError(f"'{value}' invalid for enum")
        return value

    cdef _json_decode(self, value):
        if value not in self.symbol_indexes:
            raise ValueError(f"'{value}' invalid for enum")
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        if self in created and not self.options.canonical_form_repeat_enum:
            return _CanonicalForm('"' + self.type + '"')
        created.add(self)
        return dict_to_canonical({
            'type': 'enum',
            'name': self.type,
            'symbols': self.symbols
        })

    cdef AvroType _for_writer(self, AvroType writer):
        if not isinstance(writer, EnumType):
            return
        cdef EnumType writer_enum = writer
        if not self.name_matches(writer_enum):
            return
        reader_symbols = set(self.symbols)
        writer_symbols = set(writer_enum.symbols)
        writer_extra_symbols = writer_symbols - reader_symbols
        if not writer_extra_symbols:  # reader knows about all possible symbols
            return self
        
        cdef PromotingEnumType new_type = writer.clone_base(PromotingEnumType)
        new_type.reader_type = self
        new_type.writer_type = writer_enum
        new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)
        new_type.unknown_symbols = writer_extra_symbols
        new_type.symbol_indexes = writer_enum.symbol_indexes
        return new_type

    cdef object resolve_default_value(self, object schema_default, str field):
        if self.options.string_types_default_unchanged:
            return schema_default
        return AvroType.resolve_default_value(self, schema_default, field)


cdef class PromotingEnumType(EnumType):
    cdef readonly set unknown_symbols
    cdef readonly EnumType reader_type
    cdef readonly EnumType writer_type

    cdef _json_decode(self, value):
        value = EnumType._json_decode(self, value)
        if value in self.unknown_symbols:
            if self.defaul_value is NO_DEFAULT:
                raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
            return self.default_value
        return value

    cdef _binary_buffer_decode(self, _Reader buffer):
        value = EnumType._binary_buffer_decode(self, buffer)
        if value is NO_DEFAULT:
            raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")
        return value


from pathlib import Path
import uuid


OBJ_MAGIC_BYTES = b'Obj\x01'
cdef const uint8_t[:] OBJ_MAGIC = OBJ_MAGIC_BYTES
OBJ_FILE_METADATA = Schema({"type": "map", "values": "bytes"}, bytes_codec='utf8')

_SLICE_TYPE = type(OBJ_MAGIC)


cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):
    if len(a) != len(b):
        return 1
    return memcmp(&a[0], &b[0], len(a))


cdef _Reader make_reader(src):
    if isinstance(src, _Reader):
        return src
    elif isinstance(src, bytes):
        return MemoryReader(src)
    elif isinstance(src, (str, Path)):
        return FileReader(Path(src).open('rb'))
    elif hasattr(src, 'read'):
        return FileReader(src)
    else:
        raise NotImplementedError(f"Cannot read from '{src}'")


cdef _Writer make_writer(src):
    if isinstance(src, _Writer):
        return src
    elif isinstance(src, (str, Path)):
        return FileWriter(Path(src).open('wb'))
    elif hasattr(src, 'write'):
        return FileWriter(src)
    raise NotImplementedError(f"Cannot write to '{src}'")
    

cdef class ContainerReader:

    """
    A class for reading avro object container files.

    The container can ben used as an iterator, in which case it will yield the objects in the file in order:
    ```
    for obj in ContainerReader('file.avro'):
        print(obj)
    ```

    Arguments:
     * `src`: The source to read from. Can be a file-like object, instance of `cavro.MemoryReader`, or a path to a file (str|Path)
     * `reader_schema`: The schema to use when reading objects. If not provided, the writer schema will be used.
     * `options`: An Options object to use when constructing the writer schema. Defaults to the default options. This does not affect the `reader_schema` options.
    """

    cdef readonly object metadata
    cdef readonly const uint8_t[:] marker
    cdef readonly Schema writer_schema
    cdef readonly Schema reader_schema
    cdef readonly Schema schema
    cdef readonly str codec_name
    cdef readonly size_t objects_left_in_block
    cdef MemoryReader current_block
    cdef Codec codec
    cdef _Reader reader

    def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):
        self.reader = make_reader(src)
        cdef const uint8_t[:] header = self.reader.read_n(4)
        if viewcmp(header, OBJ_MAGIC):
            raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
        self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
        writer_options = options
        if reader_schema is not None:
            writer_options = writer_options.replace(allow_invalid_default_values=True)
        self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
        if reader_schema is None:
            self.schema = self.reader_schema = self.writer_schema
        else:
            self.reader_schema = reader_schema
            self.schema = self.reader_schema.reader_for_writer(self.writer_schema)

        codec_name = self.metadata.get('avro.codec', b'null')
        self.codec_name = codec_name.decode()
        try:
            self.codec = CODECS[codec_name]
        except KeyError:
            raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")
        self.objects_left_in_block = 0
        self.current_block = MemoryReader(empty_buffer)
        self.marker = b''

    cpdef _read_marker(self):
        value = self.reader.read_n(16)
        if not len(self.marker):
            self.marker = value
        else:
            if viewcmp(value, self.marker):
                raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")

    cdef int next_block(self) except -1:
        try:
            self._read_marker()
            self.objects_left_in_block = zigzag_decode_long(self.reader)
        except (EOFError, ValueError) as e:
            raise StopIteration() from e
        cdef size_t block_size = zigzag_decode_long(self.reader)
        cdef const uint8_t[:] block_bytes
        if block_size > 0:
            block_bytes = self.codec.read_block(self.reader, block_size)
        else:
            block_bytes = empty_buffer
        self.current_block._reset_to(block_bytes)

    cpdef object next_object(self):
        while self.objects_left_in_block < 1:
            self.next_block()
        self.objects_left_in_block -= 1
        return self.schema.binary_read(self.current_block)

    def __iter__(self):
        return self

    def __next__(self):
        return self.next_object()


@cython.no_gc_clear
cdef class ContainerWriter:

    """
    A class for writing avro object container files.

    The writer can be used as a context manager, in which case it will be closed when the context exits:
    ```
    with ContainerWriter('file.avro', schema) as writer:
        writer.write_one(obj)
    ```

    Arguments:
     * `dest`: The destination to write to. Can be a file-like object, instance of `cavro.MemoryWriter`, or a path to a file (str|Path)
     * `schema`: The schema of the objects to be written.
     * `codec`: The codec to use. Must be one of the supported codecs. Default to `null`
     * `max_blocksize`: The maximum size of a block. Defaults to `16352`.
     * `write_header`: Whether to write the avro header to the file before writing blocks. Defaults to `True`.
     * `metadata`: A dictionary of metadata to write to the file. Defaults to an empty dictionary.
     * `marker`: A 16-byte marker to use to separate blocks. Defaults to a random UUID.
     * `options`: An Options object to use when writing. Defaults to the default options.
    """

    cdef _Writer writer
    cdef MemoryWriter pending_block
    cdef MemoryWriter next_block
    cdef MemoryWriter next_item

    cdef readonly Schema schema
    cdef readonly Codec codec
    cdef readonly const uint8_t[:] marker
    cdef readonly size_t max_blocksize
    cdef readonly Options options

    cdef readonly bint should_write_header

    cdef readonly size_t num_pending
    cdef readonly int blocks_written
    
    cdef readonly dict metadata

    def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
        if schema is None:
            raise ValueError('Schema is required')
        self.should_write_header = write_header
        self.num_pending = 0
        self.pending_block = MemoryWriter(max_blocksize)
        self.next_item = MemoryWriter()
        self.next_block = MemoryWriter()
        self.options = options
        
        if metadata is None:
            metadata = {}
        self.metadata = metadata

        self.writer = make_writer(dest)
        try:
            self.schema = schema
            codec_b = codec.encode('utf8')
            try:
                self.codec = CODECS[codec_b]
            except KeyError as e:
                raise CodecUnavailable(codec) from e
            if marker is None:
                marker = uuid.uuid4().bytes
            else:
                if not isinstance(marker, (bytes, _SLICE_TYPE)):
                    raise ValueError(f'Marker must be bytes, got: {type(marker)}')
                if len(marker) != 16:
                    raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
            self.marker = marker
            self.max_blocksize = max_blocksize
            self.blocks_written = 0
        except:
            self.writer = None
            raise

    def __enter__(self):
        if self.writer is None:
            raise ValueError('Container is closed')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.writer is not None:
            self.close()
        return False

    def __del__(self):
        if self.writer is not None:
            try:
                self.close()
            except ValueError as e:
                warnings.warn(f'Error closing container file during __del__: {e}', ResourceWarning)
                pass # We might be shutting down, and our writer might have been closed

    @property
    def closed(self):
        return self.writer is None

    cdef int _write_header(self) except -1:
        self.writer.write_n(OBJ_MAGIC)
        meta = self.metadata.copy()
        meta['avro.schema'] = json.dumps(self.schema.schema)
        meta['avro.codec'] = self.codec.name
        OBJ_FILE_METADATA.binary_write(self.writer, meta)
        self.writer.write_n(self.marker)
        return 0

    def close(self):
        if self.writer is None:
            raise ValueError('Trying to close a closed Container')
        cdef _Writer writer = self.writer
        self._flush_block(self.blocks_written == 0)
        self.writer = None
        self.next_item = None

    cdef int _flush_block(self, int force=False) except -1:
        cdef size_t pending_len = self.pending_block.len
        if force or self.num_pending > 0:
            if self.blocks_written == 0 and self.should_write_header:
                self._write_header()
            zigzag_encode_long(self.writer, self.num_pending)
            self.next_block.reset()
            self.codec.write_block(self.next_block, self.pending_block.view())
            zigzag_encode_long(self.writer, self.next_block.len)
            if self.next_block.len > 0:
                self.writer.write_n(self.next_block.view())
            self.writer.write_n(self.marker)
            self.blocks_written += 1
            self.pending_block.reset()
            self.num_pending = 0
        self.writer.flush()
        return 0

    def flush(self, force=False):
        self._flush_block(force)

    cpdef int write_one(self, obj) except -1:
        if self.next_item is None:
            raise ValueError('Trying to write to closed Container')
        self.next_item.reset()
        self.schema.binary_write(self.next_item, obj)

        if not self.options.container_fill_blocks:
            if self.pending_block.len + self.next_item.len > self.max_blocksize:
                self._flush_block()

        self.pending_block.write_n(self.next_item.view())
        self.num_pending += 1

        if self.options.container_fill_blocks:
            if self.pending_block.len >= self.max_blocksize:
                self._flush_block()

    def write_many(self, objs):
        for obj in objs:
            self.write_one(obj)



import io


cdef class _Writer:

    cdef int write_u8(self, uint8_t val) except -1:
        raise NotImplementedError(
            f"{type(self).__name__} does not implement write_u8")

    cdef int write_n(self, const uint8_t[:] data) except -1:
        raise NotImplementedError(
            f"{type(self).__name__} does not implement write_n")

    cdef int flush(self) except -1:
        pass


cdef class _Reader:

    cdef uint8_t read_u8(self) except? 0xba:
        raise NotImplementedError(
            f"{type(self).__name__} does not implement read_u8")

    cdef bytes read_bytes(self, Py_ssize_t n):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement read_bytes")

    cdef const uint8_t[:] read_n(self, Py_ssize_t n):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement read_n")


cdef class FileReader(_Reader):

    """
    A cavro wrapper for reading data from a file-like object (Object that implements `.read(n)`).
    
    This class will not close or seek the underlying file object

    Arguments:
        `file_obj`: The file-like object to read from.
    """

    cdef object file_obj

    def __init__(self, file_obj):
        self.file_obj = file_obj

    cdef uint8_t read_u8(self) except? 0xba:
        cdef bytes data = self.file_obj.read(1)
        if len(data) != 1:
            raise EOFError(f"End of file found trying to read 1 byte")
        return ord(data)

    cdef bytes read_bytes(self, Py_ssize_t n):
        cdef bytes result = self.file_obj.read(n)
        if len(result) != n:
            raise EOFError(f"End of file found trying to read {n} bytes")
        return result

    cdef const uint8_t[:] read_n(self, Py_ssize_t n):
        cdef bytes result = self.file_obj.read(n)
        if len(result) != n:
            raise EOFError(f"End of file found trying to read {n} bytes")
        return result


cdef class FileWriter(_Writer):
    """
    A cavro wrapper for writing data to a file-like object (Object that implements `.write(data)` and `.flush()`).
    
    This class will not close or seek the underlying file object
    """

    cdef object file_obj

    def __init__(self, file_obj):
        self.file_obj = file_obj

    cdef int write_u8(self, uint8_t val) except -1:
        self.file_obj.write(bytes([val]))

    cdef int write_n(self, const uint8_t[:] data) except -1:
        self.file_obj.write(data)

    cdef int flush(self) except -1:
        self.file_obj.flush()



@cython.final
cdef class BytesType(AvroType):
    """The avro bytes type"""
    type_name = "bytes"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        value = self.convert_value(value)
        cdef Py_ssize_t length = len(value)
        zigzag_encode_long(buffer, length)
        if length:
            buffer.write_n(value)

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef uint64_t length = zigzag_decode_long(buffer)
        return buffer.read_bytes(length)

    cdef int _get_value_fitness(self, value) except -1:
        if isinstance(value, (bytes, bytearray)):  # If bytes, we're good
            return FIT_EXACT
        # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
        if not self.options.bytes_codec:
            return FIT_NONE
        # We may be allowed to coerce to str
        MAX_FIT = FIT_OK
        if not isinstance(value, str):
            MAX_FIT = FIT_POOR
            if self.options.coerce_values_to_str:
                value = str(value)
            else:
                return FIT_NONE # We can't get a str

        # We now know that value is str and we have a codec       
        try:
            value.encode(self.options.bytes_codec)
        except (TypeError, ValueError, UnicodeEncodeError):
            return FIT_NONE
        return MAX_FIT

    cdef _json_format(self, value):
        value = self.convert_value(value)
        return value.decode('latin-1')

    cdef _json_decode(self, value):
        cdef str sval = value
        return sval.encode('latin-1')

    cpdef object _convert_value(self, object value):
        if isinstance(value, (bytes, bytearray)):
            return value
        codec = self.options.bytes_codec
        if not codec:
            raise InvalidValue(value, self)
        if not isinstance(value, str):
            if self.options.coerce_values_to_str:
                value = str(value)
            raise InvalidValue(value, self)
        # We have a str, and codec
        return value.encode(codec)

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"bytes"')

    cdef AvroType _for_writer(self, AvroType writer):
        if isinstance(writer, StringType):
            return self # Decoding is identical for bytes / str

    cdef object resolve_default_value(self, object schema_default, str field):
        if schema_default is NO_DEFAULT:
            return NO_DEFAULT
        if self.options.string_types_default_unchanged:
            return schema_default
        if self.options.bytes_default_value_utf8:
            try:
                schema_default = schema_default.encode('utf-8')
            except (AttributeError, TypeError, UnicodeEncodeError) as e:
                raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
            return schema_default
        return AvroType.resolve_default_value(self, schema_default, field)
        


@cython.final
cdef class StringType(AvroType):
    """The avro string type"""
    type_name = "string"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        value = self._convert_value(value).encode('utf8')
        cdef size_t length = len(value)
        zigzag_encode_long(buffer, length)
        if length:
            buffer.write_n(value)

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef uint64_t length = zigzag_decode_long(buffer)
        return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)

    cdef int _get_value_fitness(self, value) except -1:
        if isinstance(value, str):
            return FIT_EXACT
        if self.options.coerce_values_to_str:
            if isinstance(value, bytes):
                return FIT_OK
            return FIT_POOR
        return FIT_NONE

    cdef _json_format(self, value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        cdef str sval = value
        return sval

    cpdef object _convert_value(self, object value):
        if isinstance(value, str):
            return value
        if not self.options.coerce_values_to_str:
            raise InvalidValue(value, self)
        if isinstance(value, bytes):
            return value.decode(errors=self.options.unicode_errors)
        return str(value)

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"string"')

    cdef AvroType _for_writer(self, AvroType writer):
        if isinstance(writer, BytesType):
            return self


@cython.final
cdef class FixedType(_NamedType):
    """The avro fixed type"""
    type_name = "fixed"

    cdef readonly Py_ssize_t size

    def __init__(self, schema, source, namespace):
        self.size = source['size']
        if self.size < 0:
            raise ValueError(f'Invalid negative size for fixed: {self.size}')
        super().__init__(schema, source, namespace)

    cpdef AvroType copy(self):
        cdef FixedType new_inst =  self.clone_base()
        new_inst.size = self.size
        return new_inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})

    cpdef dict _get_schema_extra(self, set created):
        return dict(_NamedType._get_schema_extra(self, created), size=self.size)

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        value = self._convert_value(value)
        cdef Py_ssize_t length = len(value)
        if length != self.size:
            raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
        buffer.write_n(value)

    cdef _binary_buffer_decode(self, _Reader buffer):
        return bytes(buffer.read_n(self.size))

    cdef int _get_value_fitness(self, value) except -1:
        MAX_FIT = FIT_EXACT
        if not isinstance(value, (bytes, bytearray)):
            MAX_FIT = FIT_OK
            # Let's see if we can recover this
            if self.options.fixed_codec is None:
                return FIT_NONE # No codec, so can't encode

            # It's not even a str, but can we coerce it?
            if not isinstance(value, str):
                if not self.options.coerce_values_to_str: # Not allowed
                    return FIT_NONE
                MAX_FIT = FIT_POOR
                try:
                    value = str(value)
                except (TypeError, ValueError, UnicodeEncodeError):
                    return FIT_NONE
            
            # We have a str, and a codec
            try:
                value = value.encode(self.options.fixed_codec)
            except (TypeError, ValueError, UnicodeEncodeError):
                return FIT_NONE
        # We have bytes now, check the length
        cdef Py_ssize_t length = len(value)
        if length == self.size:
            return FIT_EXACT
        if length < self.size and self.options.null_pad_fixed:
            return FIT_OK
        if self.options.truncate_fixed:
            return FIT_OK
        return FIT_NONE

    cdef _json_format(self, value):
        value = self._convert_value(value)
        return value.decode('latin1', errors=self.options.unicode_errors)

    cdef _json_decode(self, value):
        cdef str sval = value
        cdef bytes encoded = sval.encode('latin1')
        if len(encoded) != self.size:
            raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")
        return encoded

    cpdef object _convert_value(self, object value):
        if not isinstance(value, (bytes, bytearray)):
            # Let's see if we can recover this
            if self.options.fixed_codec is None:
                raise ValueError(f"Invalid non-bytes value for fixed: '{value}'")

            # It's not even a str, but can we coerce it?
            if not isinstance(value, str):
                if not self.options.coerce_values_to_str: # Not allowed
                    raise ValueError(f"Invalid non-str value for fixed: '{value}'")
                value = str(value)
                    
            value = value.encode(self.options.fixed_codec)
        # We have bytes now, check the length
        cdef Py_ssize_t length = len(value)
        if length == self.size:
            return value
        if length < self.size and self.options.null_pad_fixed:
            return value + b'\x00' * (self.size - length)
        if self.options.truncate_fixed:
            return value[:self.size]
        raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")

    cdef _CanonicalForm canonical_form(self, set created):
        if self in created and not self.options.canonical_form_repeat_fixed:
            return _CanonicalForm(f'"{self.type}"')
        created.add(self)
        return dict_to_canonical({
            'type': 'fixed',
            'name': self.type,
            'size': self.size
        })

    cdef AvroType _for_writer(self, AvroType writer):
        cdef FixedType writer_fixed
        if not isinstance(writer, FixedType):
            return
        writer_fixed = writer
        if not self.name_matches(writer_fixed):
            return
        if writer_fixed.size != self.size:
            return
        return self

    cdef object resolve_default_value(self, object schema_default, str field):
        if schema_default is NO_DEFAULT:
            return NO_DEFAULT
        if self.options.string_types_default_unchanged:
            return schema_default
        if self.options.bytes_default_value_utf8:
            try:
                schema_default = schema_default.encode('utf-8')
            except (AttributeError, TypeError, UnicodeEncodeError) as e:
                raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
            return schema_default
        return AvroType.resolve_default_value(self, schema_default, field)


from libc.float cimport FLT_MAX, DBL_MAX
from cpython cimport bool as py_bool
from numpy import bool_, integer, float16 as np_f16, float32 as np_f32, float64 as np_f64
import numpy as np

from math import isnan, isinf

cdef float FLOAT_INT_THRESHOLD = 0.001

@cython.final
cdef class BoolType(AvroType):
    """The avro boolean type."""
    type_name = "boolean"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef bint bool_val = self._convert_value(value)
        buffer.write_u8(bool_val)

    cdef object _binary_buffer_decode(self, _Reader buffer):
        return py_bool(buffer.read_u8())

    cdef int _get_value_fitness(self, value) except -1:
        if isinstance(value, (py_bool, bool_)):
            return FIT_EXACT
        if self.options.coerce_values_to_boolean:
            try:
                py_bool(value)
            except (ValueError, TypeError):
                return FIT_NONE
            return FIT_POOR
        return FIT_NONE

    cpdef object _convert_value(self, object value):
        if isinstance(value, (py_bool, bool_)):
            return value
        if self.options.coerce_values_to_boolean:
            return py_bool(value)
        raise InvalidValue(value, self)

    cdef _json_format(self, value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        cdef py_bool decoded = value
        return decoded

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"boolean"')


@cython.final
cdef class IntType(AvroType):
    """The avro int type."""
    type_name = "int"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        value = self._convert_value(value)
        zigzag_encode_int(buffer, value)

    cdef _binary_buffer_decode(self, _Reader buffer):
        return zigzag_decode_int(buffer)

    cdef int _get_value_fitness(self, value) except -1:
        max_fit = FIT_EXACT
        if isinstance(value, (bool_, py_bool)):
            return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
        if not isinstance(value, (int, integer)):
            if not self.options.coerce_values_to_int:
                return FIT_NONE
            try:
                new_value = int(value)
                max_fit = FIT_OK if new_value == value else FIT_POOR
            except (ValueError, TypeError):
                return FIT_NONE
            value = new_value

        if value > INT32_MAX or value < INT32_MIN:
            return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE

        return max_fit
        
    cpdef object _convert_value(self, object value):
        if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
            raise ValueError(f'{value} not valid for long')
        if not isinstance(value, int):
            if self.options.coerce_values_to_int or isinstance(value, integer):
                value = int(value)
            else:
                raise InvalidValue(value, self)
        if value > INT32_MAX:
            if self.options.clamp_int_overflow:
                return INT32_MAX 
        elif value < INT32_MIN:
            if self.options.clamp_int_overflow:
                return INT32_MIN
        else:
            return value
        raise InvalidValue(value, self)

    cdef _json_format(self, value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
            raise InvalidValue(value, self)
        if value < INT32_MIN or value > INT32_MAX:
            raise InvalidValue(value, self)
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"int"')


@cython.final
cdef class LongType(AvroType):
    """The avro long type."""
    type_name = "long"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        value = self._convert_value(value)
        zigzag_encode_long(buffer, value)

    cdef _binary_buffer_decode(self, _Reader buffer):
        return zigzag_decode_long(buffer)

    cdef int _get_value_fitness(self, value) except -1:
        max_fit = FIT_EXACT
        if isinstance(value, (bool_, py_bool)):
            return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
        if not isinstance(value, (int, integer)):
            if not self.options.coerce_values_to_int:
                return FIT_NONE
            try:
                new_value = int(value)
                max_fit = FIT_OK if new_value == value else FIT_POOR
            except (ValueError, TypeError):
                return FIT_NONE
            value = new_value

        if value > INT64_MAX or value < INT64_MIN:
            return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE

        return max_fit
        
    cpdef object _convert_value(self, object value):
        if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
            raise ValueError(f'{value} not valid for long')
        if not isinstance(value, int):
            if self.options.coerce_values_to_int or isinstance(value, integer):
                value = int(value)
            else:
                raise InvalidValue(value, self)
        if value > INT64_MAX:
            if self.options.clamp_int_overflow:
                return INT64_MAX 
        elif value < INT64_MIN:
            if self.options.clamp_int_overflow:
                return INT64_MIN
        else:
            return value
        raise InvalidValue(value, self)

    cdef json_format(self, object value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
            raise InvalidValue(value, self)
        if value < INT64_MIN or value > INT64_MAX:
            raise InvalidValue(value, self)
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"long"')

    cdef AvroType _for_writer(self, AvroType writer):
        if isinstance(writer, IntType):
            return writer # Int is a subset of Long


@cython.final
cdef class FloatType(AvroType):
    """The avro float type."""
    type_name = "float"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef float float_val = self._convert_value(value)
        cdef uint8_t *int_val = <uint8_t*>&float_val
        buffer.write_n(int_val[:4])

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef const uint8_t[:] val = buffer.read_n(4)
        return (<float*>(&val[0]))[0]

    cdef int _get_value_fitness(self, value) except -1:
        max_fit = FIT_EXACT
        if not isinstance(value, (float, np_f16, np_f32)):
            max_fit = FIT_OK
            if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
                try:
                    value = float(value)
                except OverflowError:
                    return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
            else:
                max_fit = FIT_POOR
                if not self.options.coerce_values_to_float:
                    return FIT_NONE
                try:
                    value = float(value)
                except (ValueError, TypeError, OverflowError):
                    return FIT_NONE
        
        if self.options.clamp_float_overflow or self.options.float_out_of_range_inf:
            return max_fit

        if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):
            return max_fit
        return FIT_NONE

    cpdef object _convert_value(self, object value):
        if isinstance(value, float):
            pass
        elif isinstance(value, (np_f16, np_f32)):
            value = float(value)
        else:
            if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
                value = float(value)
            else:
                if self.options.coerce_values_to_float:
                    try:
                        value = float(value)
                    except OverflowError:
                        if self.options.clamp_float_overflow:
                            return FLT_MAX
                        else:
                            raise
                else:
                    raise InvalidValue(value, self)
        
        if isnan(value):
            if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
                return 0.0
            return value
        if isinf(value):
            if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
                return FLT_MAX if value > 0 else -FLT_MAX
            return value
        if value < -FLT_MAX:
            if self.options.clamp_float_overflow:
                return -FLT_MAX
            if self.options.float_out_of_range_inf:
                return float('-inf')
            raise InvalidValue(value, self)
        elif value > FLT_MAX:
            if self.options.clamp_float_overflow:
                return FLT_MAX
            if self.options.float_out_of_range_inf:
                return float('inf')
            raise InvalidValue(value, self)
        else:
            return value

    cdef _json_format(self, value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        if isinstance(value, (py_bool, bool_)):
            raise InvalidValue(value, self)
        if value in ('nan', 'inf', '-inf'):
            return float(value)
        if not isinstance(value, (float, int)):
            raise InvalidValue(value, self)
        if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
            raise InvalidValue(value, self)
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"float"')

    cdef AvroType _for_writer(self, AvroType writer):
        if isinstance(writer, (IntType, LongType)):
            promoted = writer.clone_base()
            promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
            return promoted


@cython.final
cdef class DoubleType(AvroType):
    """The avro double type."""
    type_name = "double"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef double float_val = self._convert_value(value)
        cdef uint8_t *int_val = <uint8_t*>&float_val
        buffer.write_n(int_val[:8])

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef const uint8_t[:] val = buffer.read_n(8)
        return (<double*>(&val[0]))[0]

    cdef int _get_value_fitness(self, value) except -1:
        max_fit = FIT_EXACT
        if not isinstance(value, (float, np_f16, np_f32, np_f64)):
            max_fit = FIT_OK
            if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
                try:
                    value = float(value)
                except OverflowError:
                    return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
            else:
                max_fit = FIT_POOR
                if not self.options.coerce_values_to_float:
                    return FIT_NONE
                try:
                    value = float(value)
                except (ValueError, TypeError, OverflowError):
                    return FIT_NONE
        
        return max_fit

    cpdef object _convert_value(self, object value):
        if isinstance(value, float):
            pass
        elif isinstance(value, (np_f16, np_f32, np_f64)):
            value = float(value)
        else:
            if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
                value = float(value)
            else:
                if self.options.coerce_values_to_float:
                    try:
                        value = float(value)
                    except OverflowError:
                        if self.options.clamp_float_overflow:
                            return DBL_MAX
                        else:
                            raise
                else:
                    raise InvalidValue(value, self)
        if isnan(value):
            if self.options.clamp_float_overflow:
                return 0.0
        elif  (value):
            if self.options.clamp_float_overflow:
                return DBL_MAX
        
        return value

    cdef _json_format(self, value):
        return self._convert_value(value)

    cdef _json_decode(self, value):
        if isinstance(value, (py_bool, bool_)):
            raise InvalidValue(value, self)
        if not isinstance(value, (float, int)):
            raise InvalidValue(value, self)
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"double"')

    cdef AvroType _for_writer(self, AvroType writer):
        cdef AvroType promoted
        if isinstance(writer, FloatType):
            return writer
        if isinstance(writer, (IntType, LongType)):
            promoted = writer.clone_base()
            promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
            return promoted



from io import BytesIO
import datetime
import pytest
import cavro
import avro.io
import avro.schema
import avro.timezones


IS_DODGY_AVRO_VERSION = avro.__version__.rsplit('.', 1)[0] in ('1.12', '1.11')

@pytest.mark.parametrize('schema, value', [
    ('"long"', 1),
    ('"long"', -1),
    ('"long"', 2**32),
    ('"long"', (2**63)-1),
    ('"long"', -(2**63)),
    ('"long"', -2**32),
    ('"int"', 1),
    ('"int"', -1),
    ('"int"', (2**31)-1),
    ('"int"', -(2**31)),
    ('"float"', -(2**31)),
    ('"float"', -(2**31)),
    (
        '{"type": "int", "logicalType": "date"}',
        datetime.date(9999, 12, 31),
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        datetime.datetime(2014, 1, 1, 1, 1, 1, 500, tzinfo=avro.timezones.utc),
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        datetime.datetime(9999, 12, 31, 23, 59, 59, 999999, tzinfo=avro.timezones.utc),
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        
        datetime.datetime(2000, 1, 18, 2, 2, 1, 123499, tzinfo=avro.timezones.tst),
    ),  
    (
        '{"type": "long", "logicalType": "timestamp-millis"}',
        datetime.datetime(9999, 12, 31, 23, 59, 59, 999999, tzinfo=avro.timezones.utc),
    ),
])
def test_encoding(schema, value):
    cavro_schema = cavro.Schema(schema, alternate_timestamp_millis_encoding=IS_DODGY_AVRO_VERSION)
    avro_schema = avro.schema.parse(schema)
    cavro_encoded = cavro_schema.binary_encode(value)

    buf = BytesIO()
    avro_encoder = avro.io.BinaryEncoder(buf)
    avro_writer = avro.io.DatumWriter(avro_schema)
    avro_writer.write(value, avro_encoder)
    avro_encoded = buf.getvalue()

    assert cavro_encoded == avro_encoded


def test_sus_default_value_behaviour():
    reader_src = '''{
        "type": "record",
        "name": "Test",
        "fields": [
            {
                "name": "H",
                "type": "bytes",
                "default": "\u00ff\u00ff"
            }
        ]
    }'''
    writer_src = '''{
        "type": "record",
        "name": "Test",
        "fields": [
            {"name": "A", "type": "int"}
        ]
    }'''
    cavro_default_reader = cavro.Schema(reader_src, record_decodes_to_dict=True)
    cavro_reader = cavro.Schema(reader_src, record_decodes_to_dict=True, bytes_default_value_utf8=True)
    cavro_writer = cavro.Schema(writer_src)
    avro_reader = avro.schema.parse(reader_src)
    avro_writer = avro.schema.parse(writer_src)
    encoded = cavro_writer.binary_encode({"A": 1})
    reader_writer = cavro_reader.reader_for_writer(cavro_writer)
    default_reader_writer = cavro_default_reader.reader_for_writer(cavro_writer)
    avro_buf = BytesIO(encoded)
    avro_decoder = avro.io.BinaryDecoder(avro_buf)
    avro_reader = avro.io.DatumReader(avro_writer, avro_reader)
    avro_decoded = avro_reader.read(avro_decoder)

    # decoded should be {"H": b"\xff\xff"}:
    assert default_reader_writer.binary_decode(encoded) == {'H': b'\xff\xff'}

    if IS_DODGY_AVRO_VERSION:
        # but avro some library versions utf-8 encode default bytes values, so it ends up being...
        assert avro_decoded == {'H': b'\xc3\xbf\xc3\xbf'}
    assert reader_writer.binary_decode(encoded) == {'H': b'\xc3\xbf\xc3\xbf'}




cimport cython
from libc.stdint cimport *

cdef int64_t RABIN_EMPTY = 0xc15d213aa4d7a795L

cdef int64_t RABIN_TABLE[256]
cdef bint RABIN_TABLE_CONFIGURED = 0


cdef init_rabin_table():
    global RABIN_TABLE_CONFIGURED
    cdef size_t i, j
    cdef int64_t mask;
    cdef int64_t value
    for i in range(256):
        value = i
        for j in range(8):
            mask = -(value & 0x1);
            value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));
        RABIN_TABLE[i] = value
    RABIN_TABLE_CONFIGURED = 1


@cython.final
cdef class Rabin:

    """
    An implementation of the 64-bit Rabin hash function  as described in the avro specification.

    The interface in this class roughly approximates the `hashlib.hash` objects.
    """

    name = 'rabin'
    digest_size = 8
    block_size = 8

    cdef readonly int64_t value

    def __init__(self, value=RABIN_EMPTY):
        if not RABIN_TABLE_CONFIGURED:
            init_rabin_table()
        self.value = value
    
    cpdef update(self, bytes data):
        cdef int64_t swizzle
        cdef int64_t shifted
        for char in data:
            swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]
            shifted = <int64_t>(<uint64_t>self.value >> 8)
            self.value = shifted ^ swizzle

    def digest(self):
        cdef uint64_t value = self.value
        return int(value).to_bytes(8, 'little', signed=False)

    def hexdigest(self):
        return self.digest().hex()

    def copy(self):
        new_ver = Rabin()
        new_ver.value = self.value
        return new_ver



import dataclasses
import cavro
import pytest


@pytest.mark.parametrize(
    ['schema', 'expected'],
    [
        ('"null"', '"null"'),
        ({'type': 'null'}, '"null"'),
        ({'type': 'boolean'}, '"boolean"'),
        ({'type': 'int'}, '"int"'),
        ({'type': 'long'}, '"long"'),
        ({'type': 'float'}, '"float"'),
        ({'type': 'double'}, '"double"'),
        ('"bytes"', '"bytes"'),
        ('"string"', '"string"'),
        (
            {'type': 'fixed', 'size': 12, 'name': 'F'},
            '{"name":"F","type":"fixed","size":12}'
        ),
        ({'type': 'enum', 'symbols': ['X', 'A', '\u00a3'], 'name': 'Foo', 'doc': 'bar'},
         '{"name":"Foo","type":"enum","symbols":["X","A","£"]}'),
        ('{"type": "enum", "symbols": ["X", "A", "\\u00a3"], "name": "Foo", "doc": "bar"}',
         '{"name":"Foo","type":"enum","symbols":["X","A","£"]}'),
        ('{"type": "record", "fields": [{"name": "a", "type": "Foo"}], "name": "Foo"}',
         '{"name":"Foo","type":"record","fields":[{"name":"a","type":"Foo"}]}'),
        ('''{"type": "record", "name": "rec", "fields": [
                {"name": "a", "type": {"type": "enum", "symbols": ["a"], "name": "X"}},
                {"name": "b", "type": {"type": "X"}}
        ]}''',
        '{"name":"rec","type":"record","fields":[{"name":"a","type":{"name":"X","type":"enum","symbols":["a"]}},{"name":"b","type":"X"}]}'),
        ({'type': 'array', 'items': 'long'}, '{"type":"array","items":"long"}'),
        (
            {'type': 'map', 'values': {'type': 'fixed', 'size': 8, 'name': 'Xxx', 'namespace': 'com.x'}},
            '{"type":"map","values":{"name":"com.x.Xxx","type":"fixed","size":8}}'
        )
    ]
)
def test_canonical_form(schema, expected):
    schema_obj = cavro.Schema(schema, cavro.PERMISSIVE_OPTIONS.replace(canonical_form_repeat_fixed=False, canonical_form_repeat_enum=False))
    assert schema_obj.canonical_form == expected


def test_default_canonical_form_repeats_fixed_enum():
    '''This appears to be a bug in avro/python, fixed types are repeated in the canonical form'''
    schema = cavro.Schema({
        'type': 'record',
        'name': 'Foo',
        'fields': [
            {'name': 'a', 'type': {'type': 'fixed', 'size': 8, 'name': 'Xxx', 'namespace': 'com.x'}},
            {'name': 'b', 'type': {'type': 'com.x.Xxx'}},
            {'name': 'c', 'type': {'type': 'enum', 'symbols': ['a'], 'name': 'X'}},
            {'name': 'd', 'type': {'type': 'X'}},
            {'name': 'e', 'type': {'type': 'Foo'}},
        ],
    },
    canonical_form_repeat_fixed=True, canonical_form_repeat_enum=True)
    assert schema.canonical_form == '''{
        "name": "Foo",
        "type": "record",
        "fields": [
            {
                "name": "a",
                "type": {
                    "name": "com.x.Xxx",
                    "type": "fixed",
                    "size": 8
                }
            },
            {
                "name": "b",
                "type": {
                    "name": "com.x.Xxx",
                    "type": "fixed",
                    "size": 8
                }
            },
            {"name": "c", "type": {"name": "X", "type": "enum", "symbols": ["a"]}},
            {"name": "d", "type": {"name": "X", "type": "enum", "symbols": ["a"]}},
            {"name": "e", "type": "Foo"}
        ]
    }'''.replace(' ', '').replace('\n', '')



cpdef str _ubin(val, n_bytes):
    cdef str raw = bin(val)[2:]
    padded = raw.rjust(n_bytes * 8, '0')
    parts = [padded[i:i+8] for i in range(0, len(padded), 8)]
    return " ".join(parts)

cdef struct random_t:
    uint64_t state
    uint64_t inc

cdef uint32_t rand(random_t *rng):
    cdef uint64_t oldstate = rng.state;
    rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
    cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
    cdef uint32_t rot = oldstate >> 59u;
    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));

cdef bytes boring_varint_encoder(uint64_t val):
    cdef uint8_t cur = 0
    out = []
    while val:
        cur = val & 0x7f
        val >>= 7
        if val:
            out.append(cur | 0x80)
    out.append(cur)
    return bytes(out)

@_tests
def _tests(add):

    def test_readvarint(uint32_t mask):
        cdef random_t state

        cdef uint32_t actual
        cdef uint32_t given
        cdef bytes src
        cdef MemoryReader reader

        cdef uint32_t i

        for i in range(50_000):
            given = rand(&state) & mask
            src = boring_varint_encoder(given)
            reader = MemoryReader(src)
            actual = read_varint(reader)
            if actual != given:
                raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
    for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:
        add(test_readvarint, mask)

    def test_zigzag_long(uint64_t mask):
        cdef random_t state

        cdef int64_t actual
        cdef int64_t given
        cdef MemoryWriter writer = MemoryWriter()
        cdef MemoryReader reader

        cdef uint64_t i
        cdef uint64_t pad

        pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
        for i in range(50_000):
            writer.reset()
            given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)
            zigzag_encode_long(writer, given)
            reader = MemoryReader(writer.bytes())
            actual = zigzag_decode_long(reader)
            if actual != given:
                raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")

    def test_readvarlong(uint64_t mask):
        cdef random_t state

        cdef uint64_t actual
        cdef uint64_t given
        cdef bytes src
        cdef MemoryReader reader

        cdef uint64_t i
        cdef uint64_t pad

        pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
        for i in range(50_000):
            given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
            src = boring_varint_encoder(given)
            src += (<char*>&pad)[:8]
            reader = MemoryReader(src)
            actual = read_varlong(reader)
            if actual != given:
                raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")

    for mask in [
        0x7f, 0x7fff, 0x7fffff, 0x7fffffff, 0x7fffffffffull,
         0x7fffffffffffull, 0x7fffffffffffffull, 0x7fffffffffffffffull,
        UINT64_MAX]:
        add(test_readvarlong, mask)
        add(test_zigzag_long, mask)




cdef class PromoteToFloat(ValueAdapter):

    """
    A value adapter that converts a value to a float on read.
    """

    cdef encode_value(self, value):
        raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")

    cdef decode_value(self, value):
        return float(value)


cdef class PromoteBytesToString(ValueAdapter):

    """
    A value adapter that decodes bytes to a string (utf8) on read.
    """

    cdef encode_value(self, value):
        raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")

    cdef decode_value(self, value):
        return value.decode("utf-8")


cdef class PromoteStringToBytes(ValueAdapter):

    """
    A value adapter that encodes a string to bytes (utf8) on read.
    """

    cdef encode_value(self, value):
        raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")

    cdef decode_value(self, value):
        return value.encode("utf-8")


cdef class CannotPromote(ValueAdapter):

    """
    A captured schema promotion error that has been deferred by `Options`, the first time this value is read, the error will be raised.
    """


    cdef readonly AvroType reader_type
    cdef readonly AvroType writer_type
    cdef readonly object extra

    def __init__(self, reader_type, writer_type, extra=None):
        self.reader_type = reader_type
        self.writer_type = writer_type
        self.extra = extra

    cdef encode_value(self, value):
        raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")

    cdef decode_value(self, value):
        raise CannotPromoteError(self.reader_type, self.writer_type, self.extra)


import cavro
import pytest

def test_nested_value_error():
    schema = cavro.Schema(
        {"type": "record", "name": "A", "fields": [
            {"name": "a", "type": {
                "type": "map", "values": {
                    "type": "array", "items": {
                        "type": "record", "name": "B", "fields": [
                            {"name": "b", "type": "int"},
                        ]
                    }
                }
            }}
        ]}
    )
    with pytest.raises(cavro.InvalidValue) as exc:
        schema.binary_encode({"a": {"x": [{"b": '1'}]}})
    assert exc.value.schema_path == ('a', 'x', 0, 'b')



cdef class NullType(AvroType):
    """The avro null type."""
    type_name = "null"

    cpdef AvroType copy(self):
        return self.clone_base()

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type'})

    cpdef dict _get_schema_extra(self, set created):
        return {}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):
            return 0
        raise ValueError(f'{repr(value)} not compatible with NullType')

    cdef _binary_buffer_decode(self, _Reader buffer):
        return None

    cdef int _get_value_fitness(self, value) except -1:
        if value is None:
            return FIT_EXACT
        if self.options.allow_false_values_for_null and not value:
            return FIT_POOR
        if self.options.missing_values_can_be_null and value is MISSING_VALUE:
            return FIT_OK
        return FIT_NONE

    cdef _json_format(self, value):
        return None

    cdef _json_decode(self, value):
        if value is not None:
            raise ValueError(f'Expected null, got {repr(value)}')
        return None

    cpdef object _convert_value(self, object value):
        return None

    cdef _CanonicalForm canonical_form(self, set created):
        return _CanonicalForm('"null"')

    cdef bint accepts_missing_value(self):
        if self.options.missing_values_can_be_null:
            return True

    cdef object resolve_default_value(self, object schema_default, str field):
        if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:
            return MISSING_VALUE 
        return AvroType.resolve_default_value(self, schema_default, field)




cdef array.array byte_buffer_template = array.array('B', [])

@cython.no_gc_clear
@cython.final
cdef class MemoryWriter(_Writer):

    """
    A class that writes to a memory buffer. The buffer is automatically resized to fit the data.

    The underlying data is accessible through the `buffer` attribute as an `array.array` of bytes.
    """

    cdef readonly array.array buffer
    cdef readonly size_t len

    def __cinit__(self, initial_size=4096):
        self.buffer = array.clone(byte_buffer_template, initial_size, zero=True)
        self.len = 0

    cdef bytes bytes(self):
        return self.buffer.data.as_chars[:self.len]

    cdef const uint8_t[:] view(self):
        return self.buffer.data.as_uchars[:self.len]

    cdef reset(self):
        self.len = 0

    cdef int write_u8(self, uint8_t val) except -1:
        if self.buffer.ob_size - self.len < 1:
            array.resize_smart(self.buffer, self.len + 1)
        self.buffer.data.as_uchars[self.len] = val
        self.len += 1

    cdef int write_n(self, const uint8_t[:] data) except -1:
        cdef size_t num = data.shape[0]
        if num == 0:
            return 0
        if self.buffer.ob_size - self.len < num:
            array.resize_smart(self.buffer, self.len + num)
        cdef const uint8_t* data_ptr = &data[0]
        memcpy(self.buffer.data.as_chars + self.len, data_ptr, num)
        self.len += num


cdef bytes empty_buffer = b"\x00"


cdef class MemoryReader(_Reader):

    """
    A class that allows cavro to read binary data from a memory buffer.

    Arguments:
     * data: The data to read from. Can be a `bytes` object or a `memoryview`.
    """

    cdef const uint8_t[:] data
    cdef const uint8_t* ptr
    cdef const uint8_t *end_ptr

    def __init__(self, data):
        self._reset_to(data)

    cdef void _reset_to(self, const uint8_t[:] data):
        if len(data):
            self.data = data
        else:
            self.data = empty_buffer

        self.ptr = &self.data[0]
        self.end_ptr = self.ptr + len(data)

    cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:
        cdef const uint8_t *ptr = self.ptr
        if <size_t>(self.end_ptr - self.ptr) < num:
            raise EOFError("Not enough input data to read value")
        self.ptr += num
        return ptr

    cdef inline int ensure(self, size_t num) except -1:
        if <size_t>(self.end_ptr - self.ptr) < num:
            raise EOFError("Not enough input data to read value")

    cdef uint8_t read_u8(self) except? 0xba:
        return self.advance(1)[0]

    cdef const uint8_t[:] read_n(self, Py_ssize_t n):
        cdef const uint8_t* ptr = self.ptr
        self.advance(n)
        return ptr[:n]

    cdef bytes read_bytes(self, Py_ssize_t n):
        cdef const uint8_t* ptr = self.ptr
        self.advance(n)
        return ptr[:n]



import json
from pathlib import Path

import pytest

import cavro

DATA_DIR = Path(__file__).parent / 'data'

WEATHER_SCHEMA = '{"name":"test.Weather","type":"record","fields":[{"name":"station","type":"string"},{"name":"time","type":"long"},{"name":"temp","type":"int"}]}'

EXPECTED_FILE = DATA_DIR / 'weather.json'

WEATHER_EXTS = [".avro"]
if cavro.HAVE_SNAPPY:
    WEATHER_EXTS.append("-snappy.avro")

@pytest.mark.parametrize('ext', WEATHER_EXTS)
def test_weather(ext):
    container = cavro.ContainerReader((DATA_DIR / f'weather{ext}').open('rb'))
    schema = container.schema
    assert schema.canonical_form == WEATHER_SCHEMA
    records = list(schema.json_encode(r) for r in container)
    expected = EXPECTED_FILE.read_text().splitlines()
    assert records == expected




@_tests
def _tests(add):

    @add
    def _test_reader_read_u8():
        buffer = MemoryReader(b'\x01\x02\x03')
        assert buffer.read_u8() == 1
        assert buffer.read_u8() == 2
        assert buffer.read_u8() == 3
        import pytest
        with pytest.raises(EOFError):
            buffer.read_u8()

    @add
    def _test_reader_read_n():
        buffer = MemoryReader(b'abacuscounter1')
        a = bytes(buffer.read_n(6)[:6])
        assert a == b'abacus', a
        b = bytes(buffer.read_n(7)[:7])
        assert b == b'counter', b
        import pytest
        with pytest.raises(EOFError):
            buffer.read_n(2)




class Order(enum.Enum):
    """
    The order of a field in a record.
    """

    ASC = 'ascending'
    DESC = 'descending'
    IGNORE = 'ignore'
    

NO_DEFAULT = _Sentinel('NO_DEFAULT')


cdef class _PlaceholderType(AvroType):

    """
    An avro type that just provied a constant value when read.
    This is used during schema promotion where the reader has a defaulted value that the writer does not.
    """

    type_name = "placeholder"

    cdef readonly object default_value

    def __init__(self, options, default_value):
        schema = Schema('"null"', options)
        AvroType.__init__(self, schema, None, None)
        self.default_value = None if default_value is MISSING_VALUE else default_value

    cdef dict _extract_metadata(self, source):
        return {}

    cpdef dict _get_schema_extra(self, set created):
        return {}
    
    cdef _make_logical(self, schema, source):
        pass

    cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:
        raise NotImplementedError("Placeholder types cannot be encoded")

    cdef _binary_buffer_decode(self, _Reader buffer):
        return self.default_value

    cdef int _get_value_fitness(self, value) except -1:
        return FIT_NONE

    cdef _json_format(self, value):
        raise NotImplementedError("Placeholder types cannot be encoded")

    cdef _json_decode(self, value):
        return self.default_value

    cpdef object _convert_value(self, object value):
        return None

    cdef _CanonicalForm canonical_form(self, set created):
        raise NotImplementedError("Placeholder types have no canonical form")


cdef list _record_data_from_dict(RecordType record, dict data):
    cdef list field_data = [None] * len(record.fields)
    cdef Py_ssize_t index = 0
    cdef Options options = record.options

    if not options.record_allow_extra_fields:
        extra_fields = data.keys() - record.field_dict.keys()
        if extra_fields:
            extra_field, *_ = extra_fields
            raise InvalidValue('...', record, (extra_field, ))
    if not options.record_encode_use_defaults:
        missing_fields = record.field_dict.keys() - data.keys()
        if missing_fields:
            missing_field, *_ = missing_fields
            raise InvalidValue('<missing>', record, (missing_field, ))

    for field in record.fields:
        value = data.get(field.name, field.default_value)
        if value is NO_DEFAULT:
            raise ValueError(f"Field {field.name} is required for record {record.type}")
        field_data[index] = field.type.convert_value(value)
        index += 1
    return field_data


@cython.freelist(8)
cdef class Record:

    """
    An instance of a record value from a schema.
    This class should never be instantiated directly, instead it forms the base-class for `RecordType.record` classes.

    Field values can be accessed using dot notation, e.g. `record.field_name`, index notation, e.g. `record['field_name']`, or by calling `_asdict()`
    Subclasses of record have a class attribute: `Type`, which is the `RecordType` schema that the record was created from.

    Internally, records are represented as a list of values, one for each field in the record, with associated field metadata.
    Subclasses can be instantiated in the following ways:
     * `Record(data: list|tuple)`: The length of data must match the number of fields in the record, and each value should correspond to the relevant field value
     * `Record(data: dict)`: The keys of the dict must match the field names, and each value should correspond to the relevant field value
     * `Record(data: Record)`: The record must be of the same type as the subclass, or must be adaptable to the subclass (Matching name and fields)
     * `Record(**kwargs)`: Each keyword argument should correspond to a field name, and the value should correspond to the relevant field value
    """

    cdef list data

    def __init__(self, data=None, **kwargs):
        cdef dict data_dict
        cdef Record rec
        if isinstance(data, type(self)):
            rec = data
            self.data = rec.data
            return
        if isinstance(data, Record):
            if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:
                data_dict = data._asdict()
            else:
                raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
            if kwargs:
                raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
        elif data:
            if kwargs:
                raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
            data_dict = data
        else:
            data_dict = kwargs
        self.data = _record_data_from_dict(self.Type, data_dict)

    def __dir__(self):
        return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']

    def __getitem__(self, name):
        cdef dict indexes = self._field_to_index
        cdef Py_ssize_t field_index = indexes[name]
        return self.data[field_index]

    cdef _repr_children(self, remain):
        cdef Record rec
        if remain < 4:
            return "{...}"
        more = False
        key_len = sum(len(f.name) + 1 for f in self.Type.fields)
        child_remain = remain - key_len
        vals = []
        for field in self.Type.fields:
            value = self[field.name]
            if isinstance(value, Record):
                rec = value
                value = rec._repr_children(child_remain)
            else:
                value = repr(value)
            repr_val = f'{field.name}: {value}'
            vals.append(repr_val)
            remain -= len(repr_val) + 1
            if remain < 4:
                more = True
                break
        return '{' + (" ".join(vals)) + ("..." if more else "") + "}"

    def __repr__(self):
        child_reprs = {f.name: self[f.name] for f in self.Type.fields}
        child_desc = ', '.join([f''])
        return f'<Record:{self.Type.type} {self._repr_children(70)}>'

    def _asdict(self):
        """
        Returns the record as a dict, with field names as keys.
        """
        cdef dict items = {}
        cdef RecordField field
        cdef Py_ssize_t i = 0
        for field in self.Type.fields:
            data = self.data[i]
            if isinstance(data, Record):
                data = data._asdict()
            items[field.name] = data
            i += 1
        return items

    def __eq__(self, other):
        if not isinstance(other, self.Type.record):
            return False
        cdef Record other_rec = other
        return self.data == other_rec.data


@cython.final
cdef class RecordField:

    """
    Holds the metadata for a record schema field.
    This class should never be instantiated directly, instead it is created by `RecordType` when parsing a schema.
    """

    cdef readonly str name
    cdef readonly str writer_name
    cdef readonly str doc
    cdef readonly AvroType type
    cdef readonly object default_value
    cdef readonly object order
    cdef readonly frozenset aliases

    def __init__(self, schema, source, namespace):
        cdef AvroType sub_type
        self.name = source['name']
        self.writer_name = self.name
        self.doc = source.get('doc', '')
        self.type = AvroType.for_source(schema, source['type'], namespace)
        self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
        self.order = Order(source.get('order', 'ascending'))
        alias_val = source.get('aliases', [])
        if not isinstance(alias_val, (list, tuple, set)):
            raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
        self.aliases = frozenset(alias_val)

    cdef for_reader(self, str reader_name, AvroType reader_type):
        cdef RecordField cloned
        cloned = RecordField.__new__(RecordField)
        cloned.name = reader_name
        cloned.writer_name = self.writer_name
        if reader_type is not None: 
            cloned.type = reader_type.for_writer(self.type)
        else:
            cloned.type = self.type
        cloned.default_value = NO_DEFAULT # If we're reading, then no default
        return cloned

    cdef placeholder(self):
        cdef RecordField rec = RecordField.__new__(RecordField)
        rec.name = self.name
        rec.doc = ''
        rec.type = _PlaceholderType(self.type.options, self.default_value)
        rec.default_value = self.default_value
        rec.order = Order.ASC
        rec.aliases = frozenset()
        return rec

    cdef _CanonicalForm canonical_form(self, set created):
        return dict_to_canonical({
            'name': self.name,
            'type': self.type.canonical_form(created),
        })

    def get_schema(self, created):
        schema = {
            'name': self.name, 
            'type': self.type.get_schema(created), 
        }
        if self.doc:
            schema['doc'] = self.doc
        if self.aliases:
            schema['aliases'] = list(self.aliases)
        if self.order != Order.ASC:
            schema['order'] = self.order.value
        if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
            try:
                schema['default'] = self.type.json_format(self.default_value)
            except InvalidValue:
                if self.type.options.allow_invalid_default_values:
                    schema['default'] = self.default_value
                else:
                    raise
        return schema


@cython.final
cdef class _FieldAccessor:
    cdef Py_ssize_t index

    def __init__(self, index):
        self.index = index

    def __get__(self, inst, cls):
        cdef Record record = inst
        return record.data[self.index]

    def __set__(self, inst, value):
        cdef Record record = inst
        record.data[self.index] = value


cdef object _make_record_class(RecordType record_type):
    attrs = {}
    field_to_index = {}
    for i, field in enumerate(record_type.fields):
        attrs[field.name] = _FieldAccessor(i)
        field_to_index[field.name] = i

    attrs['Type'] = record_type
    attrs['__slots__'] = ()
    attrs['_field_to_index'] = field_to_index
    return type(
        record_type.name,
        (Record, ),
        attrs
    )

_EMPTY_ARRAY = array.array(SSIZE_TYPECODE.decode())


cdef class RecordType(_NamedType):

    """
    The Type that corresponds to a Record in a Schema.

    Attributes:
     * `doc` Any "doc" metadata defined in the schema
     * `fields` A tuple of `RecordField` instances, one for each field in the record
     * `record` A subclass of `Record` that can be used to instantiate records of this type
    """

    type_name = 'record'

    cdef readonly str doc
    cdef readonly tuple fields
    cdef dict field_dict
    cdef readonly type record

    cdef bint _setting_up

    def __init__(self, schema, source, namespace):
        cdef Schema schema_ = schema
        self._setting_up = True
        self.doc = source.get('doc', '')
        _NamedType.__init__(self, schema, source, namespace)
        self.fields = tuple(
            RecordField(schema_, f, self.effective_namespace) for f in source['fields']
        )
        self.field_dict = {}
        for field in self.fields:
            if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
                raise InvalidName(f'Duplicate field name: {field.name}')
            self.field_dict[field.name] = field

        n_fields = len(self.fields)
        self.record = _make_record_class(self)

        self._setting_up = False
        logical = self._make_logical(schema, source)
        if logical is not None:
            self.value_adapters = self.value_adapters + (logical,)

    cdef _make_logical(self, schema, source):
        if self._setting_up:
            return
        return AvroType._make_logical(self, schema, source)

    cpdef AvroType copy(self):
        cdef RecordType new_inst = self.clone_base()
        new_inst.doc = self.doc
        new_inst.fields = self.fields
        new_inst.field_dict = self.field_dict
        new_inst.record = self.record
        return new_inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {
            'type', 
            'name', 
            'namespace', 
            'doc', 
            'aliases', 
            'fields'
        })

    def walk_types(self, visited):
        if self in visited:
            return
        yield from super().walk_types(visited)
        for field in self.fields:
            yield from field.type.walk_types(visited)

    cpdef dict _get_schema_extra(self, set created):
        extra = super(RecordType, self)._get_schema_extra(created)
        return dict(extra, fields=[f.get_schema(created) for f in self.fields])

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef RecordField field
        cdef list rec_data
        cdef Record rec
        cdef Py_ssize_t index = 0
        if isinstance(value, Record):
            if isinstance(value, self.record):
                rec = value
            else:
                rec = self.record(value)
            rec_data = rec.data
            for field in self.fields:
                field_value = rec_data[index]
                if field_value is NO_DEFAULT:
                    raise ValueError(f"required field '{field.name}' missing")
                field.type.binary_buffer_encode(buffer, field_value)
                index = index + 1
            return 0
        if not self.options.record_can_encode_dict:
            raise InvalidValue(value, self, (self.name, ))
        if not self.options.record_allow_extra_fields:
            extra_fields = value.keys() - self.field_dict.keys()
            if extra_fields:
                extra_field, *_ = extra_fields
                raise InvalidValue('...', self, (extra_field, ))
        if not self.options.record_encode_use_defaults:
            missing_fields = self.field_dict.keys() - value.keys()
            if missing_fields:
                missing_field, *_ = missing_fields
                raise InvalidValue('<missing>', self, (missing_field, ))
        for field in self.fields:
            try:
                field_value = value.get(field.name, field.default_value)
            except AttributeError as e:
                raise InvalidValue(value, self, (self.name, )) from e
            if field_value is NO_DEFAULT:
                raise ValueError(f"required field '{field.name}' missing")
            try:
                field.type.binary_buffer_encode(buffer, field_value)
            except InvalidValue as e:
                prefix = (field.name, )
                if self.options.invalid_value_includes_record_name:
                    prefix = (self.type, ) + prefix
                e.schema_path = prefix + e.schema_path
                raise

    cdef _binary_buffer_decode_record(self, _Reader buffer):
        cdef RecordField field
        cdef list data = [None] * len(self.fields)
        cdef Record rec
        cdef Py_ssize_t index = 0
        for field in self.fields:
            data[index] = field.type.binary_buffer_decode(buffer)
            index += 1
        rec = Record.__new__(self.record)
        rec.data = data
        return rec

    cdef _binary_buffer_decode_dict(self, _Reader buffer):
        cdef RecordField field
        cdef dict data = {}
        for field in self.fields:
            value = field.type.binary_buffer_decode(buffer)
            if field.name is not None:
                data[field.name] = value
        return data

    cdef _binary_buffer_decode(self, _Reader buffer):
        if self.options.record_decodes_to_dict:
            return self._binary_buffer_decode_dict(buffer)
        return self._binary_buffer_decode_record(buffer)

    cdef int _get_value_fitness(self, value) except -1:
        cdef int level = FIT_OK
        cdef RecordField field
        if isinstance(value, self.record):
            return FIT_EXACT
        if isinstance(value, Record):
            if value.Type.name != self.type:
                return FIT_NONE
            value = value._asdict()
        if isinstance(value, dict):
            if self.options.record_values_type_hint and '-type' in value:
                value = value.copy()
                type_val = value.pop('-type')
                if type_val != self.type:
                    return FIT_NONE
            remaining_keys = set(value.keys())
            for field in self.fields:
                field_value = value.get(field.name, field.default_value)
                if field_value is NO_DEFAULT:
                    return FIT_NONE
                level = min(level, field.type.get_value_fitness(field_value))
                if level <= FIT_NONE:
                    return level
                remaining_keys.discard(field.name)
            if remaining_keys:
                if self.options.record_allow_extra_fields:
                    return FIT_POOR
                return FIT_NONE
            return level

    cdef _json_format(self, value):
        cdef Record record = self._convert_value(value)
        cdef AvroType field_type
        out = {}
        for field, value in zip(self.fields, record.data):
            field_type = field.type
            out[field.name] = field_type.json_format(value)
        return out

    cdef json_decode_record(self, dict value):
        cdef list data = [None] * len(self.fields)
        cdef RecordField field
        cdef Record rec
        cdef Py_ssize_t index = 0

        for field in self.fields:
            field_value = value.get(field.writer_name, MISSING_VALUE)
            if field_value is MISSING_VALUE:
                if field.default_value is NO_DEFAULT:
                    raise ValueError(f"required field '{field.name}' missing")
                field_value = field.default_value
            else:
                field_value = field.type.json_decode(field_value)
            data[index] = field_value
            index += 1
        rec = Record.__new__(self.record)
        rec.data = data
        return rec

    cdef json_decode_dict(self, dict value):
        cdef dict data = {}
        cdef RecordField field

        for field in self.fields:
            field_value = value.get(field.writer_name, MISSING_VALUE)
            if field_value is MISSING_VALUE:
                if field.default_value is NO_DEFAULT:
                    raise ValueError(f"required field '{field.name}' missing")
                field_value = field.default_value
            else:
                field_value = field.type.json_decode(field_value)
            data[field.name] = field_value
        return data

    cdef _json_decode(self, value):
        if self.options.record_decodes_to_dict:
            return self.json_decode_dict(value)
        return self.json_decode_record(value)

    cpdef object _convert_value(self, object value):
        if isinstance(value, self.record):
            return value
        return self.record(value)

    cdef _CanonicalForm canonical_form(self, set created):
        cdef RecordField field
        if self in created:
            return _CanonicalForm('"' + self.type + '"')
        created.add(self)
        return dict_to_canonical({
            'type': 'record',
            'name': self.type,
            'fields': [field.canonical_form(created) for field in self.fields]
        })

    cdef AvroType _for_writer(self, AvroType writer):
        cdef PromotingRecordType cloned
        cdef RecordType writer_rec
        cdef RecordField field
        cdef Py_ssize_t i
        if not isinstance(writer, RecordType):
            return
        writer_rec = writer
        if not self.name_matches(writer_rec):
            return
        # This allows us to go from the field name to the index in the reader record
        reader_field_idx = {}
        reader_fields = {}
        aliases = {}

        i = 0
        for field in self.fields:
            reader_field_idx[field.name] = i
            reader_fields[i] = field
            for alias in field.aliases:
                aliases[alias] = i
            i += 1

        # We're going to have to read each writer field in order anyway, but if it's a reader field,
        # Then it ends up in the corresponding decode_index slot, but should also be promoted etc..
        out_decode_indexes = []
        out_fields = []
        decoded_reader_fields = set()

        # Todo, only use promoting type if the field order/naming is different
        for field in  writer_rec.fields:
            if field.name in reader_field_idx:
                reader_index = reader_field_idx[field.name]
                reader_field = reader_fields[reader_index]
            elif field.name in aliases:
                reader_index = aliases[field.name]
                reader_field = reader_fields[reader_index]
            else:
                # This field is not in the reader, so we read it and discard (index = -1)
                out_decode_indexes.append(-1)
                out_fields.append(field.for_reader(None, None))
                continue
                
            out_decode_indexes.append(reader_index)
            out_fields.append(field.for_reader(reader_field.name, reader_field.type))
            decoded_reader_fields.add(reader_index)
        
        extra_reader_fields = reader_fields.keys() - decoded_reader_fields
        for reader_index in extra_reader_fields:
            field = reader_fields[reader_index]

            if field.default_value is NO_DEFAULT:
                raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")
            out_decode_indexes.append(reader_index)
            out_fields.append(field.placeholder())
        
        cloned = self.clone_base(PromotingRecordType)
        cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)
        for i, idx in enumerate(out_decode_indexes):
            cloned.decode_indexes[i] = idx
        cloned.fields = tuple(out_fields)
        cloned.field_dict = {f.name: f for f in cloned.fields}
        cloned.record = self.record
        return cloned


cdef class PromotingRecordType(RecordType):

    """
    A variant of a `RecordType`, specialized for reading records from a different schema from a writer.
    """

    cdef Py_ssize_t [:] decode_indexes

    cdef _binary_buffer_decode_record(self, _Reader buffer):
        cdef RecordField field
        cdef list data = [None] * len(self.fields)
        cdef Record rec
        cdef Py_ssize_t index = 0
        cdef Py_ssize_t field_index
        for field in self.fields:
            field_index = self.decode_indexes[index]
            value = field.type.binary_buffer_decode(buffer)
            if field_index >= 0: 
                data[field_index] = value
            index += 1
        rec = Record.__new__(self.record)
        rec.data = data
        return rec



cdef int FIT_NONE = 0  # Value is never valid for type
cdef int FIT_POOR = 1  # Value may be valid depending on the schema options
cdef int FIT_OK = 2    # Value can be converted to the correct type
cdef int FIT_EXACT = 3 # Value is the exact type and needs no further conversion


CANONICAL_FORM_KEYS = ('name', 'type', 'fields', 'symbols', 'items', 'values', 'size')

cdef class _Sentinel:
    cdef readonly str name
    
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f'<{self.name}: {id(self)}>'

MISSING_VALUE = _Sentinel('MISSING_VALUE')

cdef class _CanonicalForm(str):
    pass

cdef _CanonicalForm dict_to_canonical(data):
    if isinstance(data, dict):
        pairs=[]
        for key in CANONICAL_FORM_KEYS:
            if key in data:
                pairs.append((f'"{key}"', dict_to_canonical(data[key])))
        pair_str = ','.join(':'.join(p) for p in pairs)
        return _CanonicalForm('{' + pair_str + '}')
    elif isinstance(data, (tuple, list)):
        values = ','.join([dict_to_canonical(v) for v in data])
        return _CanonicalForm('[' + values + ']')
    elif isinstance(data, _CanonicalForm):
        return data
    else:
        return _CanonicalForm(json.dumps(data, ensure_ascii=False))


cdef dict _strip_keys(dict source, set keys):
    return {k: v for k, v in source.items() if k not in keys}


cdef class AvroType:

    """
    The base class for all Avro types.
    `cavro` separates the concept of a `Schema` from a `Type`, which is not strictly neccessary, but
    makes some management of state a bit easier.
    Here, a `Type` is the specific implementation of the data management, whereas a `Schema` is the
    public interface for a schema definition, typically (but not always) containing multiple types either in unions or in nested record fields.

    For normal usage, the `AvroType` class and subclasses can largely be ignored.
    """

    type_name = NotImplemented

    cdef readonly Options options
    cdef readonly object metadata
    cdef readonly tuple[ValueAdapter] value_adapters

    @classmethod
    def for_source(cls, schema, source, namespace=None):
        cdef Schema schema_ob = schema
        if isinstance(source, (list, tuple)):
            return UnionType(schema, source, namespace)
        if isinstance(source, str):
            source = {'type': source}
        try:
            type_name = source['type']
        except (TypeError, KeyError) as e:
            raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
        inst = schema_ob.find_type(namespace, type_name, False)
        if inst is not None:
            return inst
        if type_name in TYPES_BY_NAME:
            return TYPES_BY_NAME[type_name](schema, source, namespace)
        if type_name == 'error' and schema_ob.options.allow_error_type:
            return RecordType(schema, source, namespace)
        namespaced = resolve_namespaced_name(namespace, type_name)
        raise UnknownType(namespaced)

    @classmethod
    def for_schema(cls, schema):
        return AvroType.for_source(schema, schema.source)

    def __init__(self, schema, source, namespace, value_adapters=tuple()):
        cdef ValueAdapter logical
        self.options = schema.options
        self.metadata = PyDictProxy_New(self._extract_metadata(source))
        
        logical = self._make_logical(schema, source)
        if logical is not None:
            self.value_adapters = value_adapters + (logical,)
        else:
            self.value_adapters = value_adapters

    cdef AvroType clone_base(self, cls=None):
        if cls is None:
            cls = self.__class__
        cdef AvroType clone = cls.__new__(cls)
        clone.options = self.options
        clone.metadata = self.metadata
        clone.value_adapters = self.value_adapters
        return clone

    cpdef AvroType copy(self):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement clone")

    cdef _make_logical(self, schema, source):
        cdef str logical_type_name = source.get('logicalType')
        if logical_type_name is None:
            return
        logical_type_classes = schema.logical_types.get(logical_type_name, [])
        for cls in logical_type_classes:
            inst = cls.for_type(self)
            if inst is not None:
                return inst

    property type:
        def __get__(self):
            return self.type_name

    cdef dict _extract_metadata(self, source):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _extract_metadata")

    cpdef object _convert_value(self, object value):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _convert_value")

    cpdef object convert_value(self, object value, check_value=False):
        if check_value:
            self.assert_value(value)
        return self._convert_value(value)

    cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef ValueAdapter adapter

        if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
            type_name, inner_value = value
            if type_name == self.type:
                try:
                    return self.binary_buffer_encode(buffer, inner_value)
                except Exception as e:
                    pass

        for adapter in self.value_adapters:
            value = adapter.encode_value(value)
        return self._binary_buffer_encode(buffer, value)

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _binary_buffer_encode")

    cdef binary_buffer_decode(self, _Reader buffer):
        cdef ValueAdapter adapter
        value = self._binary_buffer_decode(buffer)
        for adapter in reversed(self.value_adapters):
            value = adapter.decode_value(value)
        return value

    cdef _binary_buffer_decode(self, _Reader buffer):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement binary_buffer_decode")

    cdef int get_value_fitness(self, value) except -1:
        cdef ValueAdapter adapter
        
        if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
            type_name, inner_value = value
            if type_name == self.type:
                inner_fitness = self._get_value_fitness(inner_value)
                return inner_fitness

        for adapter in self.value_adapters:
            try:
                value = adapter.encode_value(value)
            except Exception as e:
                return FIT_NONE
        value_fit = self._get_value_fitness(value)
        if value_fit > FIT_NONE:
            return value_fit

        return FIT_NONE

    cdef int _get_value_fitness(self, value) except -1:
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _get_value_fitness")

    cdef int assert_value(self, object value) except -1:
        cdef int fitness = self.get_value_fitness(value)
        if fitness == FIT_NONE:
            raise ValueError(f"'{value}' not valid for {type(self).__name__}")

    cdef json_format(self, value):
        cdef ValueAdapter adapter

        if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
            type_name, inner_value = value
            if type_name == self.type:
                try:
                    return self.json_format(inner_value)
                except Exception as e:
                    pass

        for adapter in self.value_adapters:
            value = adapter.encode_value(value)
        return self._json_format(value)

    cdef _json_format(self, value):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _json_format")

    cdef json_decode(self, value):
        cdef ValueAdapter adapter
        value = self._json_decode(value)
        for adapter in reversed(self.value_adapters):
            value = adapter.decode_value(value)
        return value

    cdef _json_decode(self, value):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _json_decode")

    cdef _CanonicalForm canonical_form(self, set created):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement canonical_form")

    cpdef dict _get_schema_extra(self, set created):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement _get_schema_extra")

    cdef AvroType for_writer(self, AvroType writer, bint _allow_deferrals=True):
        cdef UnionType writer_union
        cdef AvroType promoted
        #cdef AvroType cloned
        cdef bint allow_deferrals = _allow_deferrals and self.options.defer_schema_promotion_errors
        
        promote_error = None
        promoted = None

        if self.canonical_form(set()) == writer.canonical_form(set()):
            return writer
        if isinstance(writer, UnionType) and not isinstance(self, UnionType):
            writer_union = writer
            try:
                promoted = writer_union.for_reader(self)
            except CannotPromoteError as e:
                promote_error = e
        else:
            try:
                promoted = self._for_writer(writer)
            except CannotPromoteError as e:
                promote_error = e
            
        if promoted is None and promote_error is None:
            promote_error = CannotPromoteError(self, writer)
        
        if promote_error is not None:
            if allow_deferrals:
                cloned = writer.copy()
                cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
                return cloned
            raise promote_error
        return promoted

    cdef AvroType _for_writer(self, AvroType writer):
        pass

    cdef bint accepts_missing_value(self):
        return False

    cdef object resolve_default_value(self, object schema_default, str field):
        if schema_default is NO_DEFAULT:
            return NO_DEFAULT
        try:
            return self.json_decode(schema_default)
        except (TypeError, AttributeError, ValueError) as e:
            if self.options.allow_invalid_default_values:
                return schema_default
            raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e

    def walk_types(self, visited):
        if self in visited:
            return
        visited.add(self)
        yield self

    cpdef get_schema(self, created=None):
        if created is None:
            created = set()
        if self.options.expand_types_in_schema:
            created = created.copy()
        if isinstance(self, _NamedType):
            type_name = self.type
            if type_name in created:
                return type_name
            else:
                created.add(type_name)

        extra = self._get_schema_extra(created)
        if self.metadata:
            extra.update(self.metadata)
        if extra:
            extra['type'] = self.type_name
            return extra
        else:
            return self.type_name

    def __str__(self):
        if self.options.types_str_to_schema:
            return json.dumps(self.get_schema())
        else:
            return super().__str__()


cdef class _NamedType(AvroType):
    """Base class for all named types"""

    cdef readonly str name
    cdef readonly str namespace
    cdef readonly str effective_namespace
    cdef readonly frozenset aliases

    def __init__(self, schema, source, parse_namespace):
        cdef Schema schema_t = schema
        cdef str effective_namespace = None
        cdef str name = source['name']
        if '.' in name:
            effective_namespace, name = name.rsplit('.', 1)

        if not schema_t.options.allow_leading_dot_in_names:
            if effective_namespace == '':
                raise InvalidName(f"The null namespace cannot be specified in name fields")

        name_pattern = schema_t.options.name_pattern
        if schema_t.options.enforce_type_name_rules:    
            if not name_pattern.fullmatch(name):
                raise InvalidName(f"Type name '{name}' is not valid")
        
        self.name = name
        self.namespace = source.get('namespace')
        if effective_namespace is None:
            effective_namespace = self.namespace
            if effective_namespace is None:
                effective_namespace = parse_namespace

        self.effective_namespace = effective_namespace or None

        if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:
            namespace_parts = self.effective_namespace.split('.')
            for part in namespace_parts:
                if not part or not name_pattern.fullmatch(part):
                    raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")

        if not schema_t.options.allow_primitive_name_collision:
            if name in PRIMITIVE_TYPES:
                if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:
                    pass
                else:
                    raise ValueError(f"'{name}' is not allowed as a name")

        alias_val = source.get('aliases', [])
        if not isinstance(alias_val, (list, tuple, set)):
            if schema_t.options.allow_aliases_to_be_string:
                alias_val = [alias_val]
            else:
                raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
        self.aliases = frozenset(alias_val)
        super().__init__(schema, source, parse_namespace)
        schema_t.register_type(self.effective_namespace, self.name, self)

    property type:
        def __get__(self):
            return resolve_namespaced_name(self.effective_namespace, self.name)

    cdef bint name_matches(self, _NamedType other):
        if self.name == other.name:
            return True
        if other.name in self.aliases:
            return True
        # Let's get creative
        if other.effective_namespace and other.type in self.aliases:
            return True
        return False

    cdef frozenset get_namespaced_aliases(self):
        cdef frozenset all_names = self.aliases | {self.name}
        if not self.effective_namespace:
            return all_names
        return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)

    cpdef dict _get_schema_extra(self, set created):
        schema = {}
        if self.effective_namespace:
            if self.options.inline_namespaces:
                schema['name'] = f"{self.effective_namespace}.{self.name}"
            else:
                schema['namespace'] = self.effective_namespace
        if self.aliases:
            schema['aliases'] = list(self.aliases)
        if 'name' not in schema:
            schema['name'] = self.name
        return schema

    cdef AvroType clone_base(self, cls=None):
        cdef _NamedType inst = AvroType.clone_base(self, cls)
        inst.name = self.name
        inst.namespace = self.namespace
        inst.effective_namespace = self.effective_namespace
        inst.aliases = self.aliases
        return inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {
            'type', 
            'name', 
            'namespace', 
            'aliases', 
        })


PRIMITIVE_TYPES = {
    'null': NullType,
    'boolean': BoolType,
    'int': IntType,
    'long': LongType,
    'float': FloatType,
    'double': DoubleType,
    'bytes': BytesType,
    'string': StringType,
}

TYPES_BY_NAME = dict(
    PRIMITIVE_TYPES,
    map=MapType,
    enum=EnumType,
    record=RecordType,
    fixed=FixedType,
    array=ArrayType,
)



class CavroException(Exception):
    """
    Base class for exceptions raised by cavro
    """


class InvalidName(CavroException):
    """
    The schema contains a type or enum symbol with an invalid name (as per the avro specification)
    """


class UnknownType(CavroException):

    """
    The schema contains an unexptected type name (either a missing named-type definition, or invalid primitive type)
    """

    def __init__(self, name):
        self.name = name
        super().__init__(f"Unknown type: {name}")


class DuplicateName(CavroException):
    """
    A record contains multiple fields with the same name, a schema contains multiple types of the same name, or an enum has multiple identical symbols.
    """


class InvalidHasher(CavroException):
    """
    An unknown hash method was requested
    """


class ExponentTooLarge(CavroException):
    """
    The exponent of a decimal value is too large to be represented in the given type
    """
    


class CodecUnavailable(CavroException):
    """
    A requested codec (or codec in a container) is not available or is unknown.
    """


class CannotPromoteError(CavroException):

    """
    A schema cannot be promoted to another schema. (reader/writer schema promotion)

    Attributes:
     * `reader_type`: The schema type of the reader
     * `writer_type`: The schema type of the writer
     * `extra`: An optional extra message
    """

    def __init__(self, reader_type, writer_type, extra=None):
        self.reader_type = reader_type
        self.writer_type = writer_type
        self.extra = extra
        msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
        if extra:
            msg += ': ' + extra
        super().__init__(msg)


class InvalidValue(CavroException, ValueError):

    """
    A value is invalid for a given avro type.
        
    Attributes:
     * `value`: The value that caused the error
     * `dest_type`: The schema type that the value was being converted to
     * `schema_path`: A sequence of identifiers (field names etc) to help locate the value that caused the error
    """
    
    def __init__(self, value, dest_type, path=()):
        self.value = value
        self.dest_type = dest_type
        self.schema_path = path
        super().__init__()

    def __str__(self):
        if self.schema_path:
            path_str = ".".join(str(p) for p in self.schema_path)
            return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
        return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"



cdef class MapType(AvroType):
    """The avro map type."""
    type_name = "map"

    cdef StringType key_type
    cdef readonly AvroType value_type

    def __init__(self, schema, source, namespace):
        super().__init__(schema, source, namespace)
        self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options
        self.value_type = AvroType.for_source(schema, source['values'], namespace)

    cpdef AvroType copy(self):
        cdef MapType new_inst = self.clone_base()
        new_inst.key_type = self.key_type
        new_inst.value_type = self.value_type
        return new_inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type', 'values'})

    def walk_types(self, visited):
        if self in visited:
            return
        yield from super().walk_types(visited)
        yield from self.value_type.walk_types(visited)

    cpdef dict _get_schema_extra(self, set created):
        return {'values': self.value_type.get_schema(created)}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        if hasattr(value, 'items'):
            value = value.items()
        zigzag_encode_long(buffer, len(value))
        if value:
            for key, item_value in value:
                self.key_type.binary_buffer_encode(buffer, key)
                try:
                    self.value_type.binary_buffer_encode(buffer, item_value)
                except InvalidValue as e:
                    if self.options.invalid_value_include_map_key:
                        e.schema_path = (key, ) + e.schema_path
                    raise
            zigzag_encode_long(buffer, 0)
        return 0

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef dict out = {}
        cdef size_t length
        cdef str key
        while True:
            length = zigzag_decode_long(buffer)
            if length == 0:
                return out
            while length:
                key = self.key_type.binary_buffer_decode(buffer)
                value = self.value_type.binary_buffer_decode(buffer)
                out[key] = value
                length -= 1

    cdef _json_format(self, value):
        cdef str key
        cdef dict out = {}
        if hasattr(value, 'items'):
            value = value.items()
        for key, item_value in value:
            out[key] = self.value_type.json_format(item_value)
        return out

    cdef _json_decode(self, value):
        cdef dict inp = value
        cdef dict out = {}
        cdef str key
        for key, val in inp.items():
            out[key] = self.value_type.json_decode(val)
        return out

    cdef int _get_value_fitness(self, value) except -1:
        cdef int level = FIT_OK
        if isinstance(value, dict):
            level = FIT_EXACT
        
        elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
            return FIT_POOR
            
        if hasattr(value, 'items'):
            value = value.items()
        try:
            it = iter(value)
        except (TypeError, KeyError):
            return False
        while True:
            try:
                item = next(it)
            except StopIteration:
                break
            except:
                return FIT_NONE
            try:
                key, item_value = item
            except (TypeError, ValueError, IndexError, KeyError):
                return FIT_NONE
            level = min(self.key_type.get_value_fitness(key), level)
            level = min(self.value_type.get_value_fitness(item_value), level)
            if level <= FIT_NONE:
                break
        return level

    cdef dict _make_converted_map(self, value):
        cdef AvroType key_type = self.key_type
        cdef AvroType value_type = self.value_type
        cdef dict out = {}
        for item in value:
            try:
                key, item_value = item
            except (TypeError, ValueError, IndexError):
                raise ValueError(f"'{value}' is not a mapping")
            out[key_type._convert_value(key)] = value_type._convert_value(item_value)
        return out

    cpdef object convert_value(self, object orig_value, check_value=True):
        cdef int key_fitness
        cdef int value_fitness
        cdef AvroType key_type = self.key_type
        cdef AvroType value_type = self.value_type

        if isinstance(orig_value, dict):
            value = orig_value.items()
        else:
            value = orig_value

        if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
            return {}

        it = iter(value)
        for item in it:
            try:
                key, item_value = item
            except (TypeError, ValueError, IndexError):
                raise ValueError(f"'{value}' is not a mapping")
            key_fitness = key_type.get_value_fitness(key)
            if key_fitness == FIT_NONE:
                raise ValueError(f"'{key}' is not a valid key for map type")
            elif key_fitness < FIT_EXACT:
                return self._make_converted_map(iter(value))

            value_fitness = value_type.get_value_fitness(item_value)
            if value_fitness == FIT_NONE:
                raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
            elif value_fitness < FIT_EXACT:
                return self._make_converted_map(iter(value))
        return orig_value

    cdef _CanonicalForm canonical_form(self, set created):
        return dict_to_canonical({
            'type': 'map',
            'values': self.value_type.canonical_form(created)
        })

    cdef AvroType _for_writer(self, AvroType writer):
        cdef MapType cloned
        cdef MapType writer_map
        if isinstance(writer, MapType):
            writer_map = writer
            promoted_value = self.value_type._for_writer(writer.value_type)
            if promoted_value is not None:
                cloned = self.clone_base()
                cloned.key_type = writer_map.key_type
                cloned.value_type = promoted_value
                return cloned

    cdef bint accepts_missing_value(self):
        if self.options .missing_values_can_be_empty_container:
            return True


import hashlib
import cavro
import pytest
import re

from pathlib import Path

PATTERN_1 = re.compile(r"""\
<<INPUT ([^\n]+)
<<canonical ([^\n]+)(?:
<<fingerprint ([^\n]+))?
""")

PATTERN_2 = re.compile(r"""\
<<INPUT
(.*?)
INPUT
<<canonical ([^\n]+)(?:
<<fingerprint ([^\n]+))?
""", re.S)

def read_tests():
    tests_file = Path(__file__).parent / 'data' / 'schema-tests.txt'
    with tests_file.open() as fh:
        source = fh.read()
    return PATTERN_1.findall(source) + PATTERN_2.findall(source)

TESTS = read_tests()

@pytest.mark.parametrize("schema_text, canonical, fingerprint", TESTS)
def test_avro_schema_tests(schema_text, canonical, fingerprint):
    schema = cavro.Schema(schema_text, cavro.PERMISSIVE_OPTIONS)
    assert schema.canonical_form == canonical
    for method in ['sha256', 'md5']:
        assert (
            schema.fingerprint(method).hexdigest() 
            == hashlib.new(method, canonical.encode()).hexdigest()
        )
    if fingerprint:
        assert schema.fingerprint().value == int(fingerprint)


def test_wrapping():
    schema = cavro.Schema('[{"type": "int"}, {"type": "string"}]')
    assert schema.schema == ['int', 'string']
    int_schema = schema._wrap_type(schema.type.union_types[0])
    assert int_schema.schema == 'int'
    str_schema = cavro.Schema._wrap_type(schema.type.union_types[1])
    assert str_schema.schema == 'string'


from libc.stdint cimport *

cdef uint32_t read_varint(_Reader buf) except? 0xfffffbad:
    cdef uint32_t cur
    cdef uint32_t shift = 0
    cdef uint32_t value = 0
    while True:
        cur = buf.read_u8()
        value |= (cur & 0b01111111) << shift
        shift += 7
        if not cur & 0b10000000:
            return value


cdef uint64_t read_varlong(_Reader buf) except? 0xfffffffffffffbadull:
    cdef uint64_t cur
    cdef uint64_t shift = 0
    cdef uint64_t value = 0
    while True:
        cur = buf.read_u8()
        value |= (cur & 0b01111111) << shift
        shift += 7
        if not cur & 0b10000000:
            return value


cdef int32_t zigzag_decode_int(_Reader buf) except? 0x7ffffbadu:
    cdef uint32_t value = read_varint(buf)
    return (value >> 1) ^ (-(value & 1))


cdef int64_t zigzag_decode_long(_Reader buf) except? 0x7ffffffffffffbadull:
    cdef uint64_t value = read_varlong(buf)
    return (value >> 1) ^ (-(value & 1ull))

@cython.cdivision(True)
cdef int zigzag_encode_int(_Writer buf, int32_t value) except -1:
    cdef uint32_t zz = (value << 1) ^ (value >> 31)
    cdef uint8_t cur
    if zz == 0:
        buf.write_u8(0)
    while zz:
        cur = zz & 0x7f
        zz >>= 7
        if zz:
            cur |= 0x80
        buf.write_u8(cur)

@cython.cdivision(True)
cdef int zigzag_encode_long(_Writer buf, int64_t value) except -1:
    cdef uint64_t zz = (value << 1) ^ (value >> 63)
    cdef uint8_t cur
    if zz == 0:
        buf.write_u8(0)
    while zz:
        cur = zz & 0b01111111
        zz >>= 7
        if zz:
            cur |= 0b10000000
        buf.write_u8(cur)




@_perf
def _perf(add):
    def _test_zigzag_perf(const uint64_t mask):
        from time import perf_counter_ns
        import os
        cdef random_t state

        cdef uint64_t BATCH_SIZE = 5_000_000

        cdef uint64_t actual
        cdef uint64_t given
        cdef bytes src
        cdef MemoryWriter writer = MemoryWriter(3442987655)
        cdef MemoryReader reader

        cdef size_t total_nums = 0
        cdef size_t n
        cdef uint64_t i
        cdef uint64_t out_state = 0
        t_a = perf_counter_ns()
        for i in range(BATCH_SIZE):
            given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
            zigzag_encode_long(writer, given)
            total_nums += 1
        write_time = perf_counter_ns() - t_a
        t_a = perf_counter_ns()
        reader = MemoryReader(writer.bytes())
        copy_time = perf_counter_ns() - t_a
        t_a = perf_counter_ns()
        for n in range(total_nums):
            out_state ^= zigzag_decode_long(reader)
        read_time = perf_counter_ns() - t_a

        write_perf = (write_time) / BATCH_SIZE
        read_perf = (read_time) / BATCH_SIZE
        print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")

    for mask in [
        0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX,
        0x7ffffffffful, 0x7ffffffffffful, 0x7ffffffffffffful,
        0x7ffffffffffffffful, UINT64_MAX
        ]:
        add(_test_zigzag_perf, mask)



import numpy
import collections

@cython.final
cdef class ArrayType(AvroType):
    """The avro array type."""
    type_name = "array"

    cdef readonly AvroType item_type

    def __init__(self, schema, source, namespace):
        super().__init__(schema, source, namespace)
        self.item_type = AvroType.for_source(schema, source['items'], namespace)

    cpdef AvroType copy(self):
        cdef ArrayType new_inst = self.clone_base()
        new_inst.item_type = self.item_type
        return new_inst

    cdef dict _extract_metadata(self, source):
        return _strip_keys(dict(source), {'type', 'items'})

    def walk_types(self, visited):
        if self in visited:
            return
        yield from super().walk_types(visited)
        yield from self.item_type.walk_types(visited)

    cpdef dict _get_schema_extra(self, set created):
        return {'items': self.item_type.get_schema(created)}

    cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
        cdef size_t idx = 0
        if len(value):
            zigzag_encode_long(buffer, len(value))
            for item in value:
                try:
                    self.item_type.binary_buffer_encode(buffer, item)
                except InvalidValue as e:
                    if self.options.invalid_value_include_array_index:
                        e.schema_path = (idx, ) + e.schema_path
                    raise
                idx += 1
        zigzag_encode_long(buffer, 0)

    cdef _binary_buffer_decode(self, _Reader buffer):
        cdef list out = []
        cdef size_t length
        cdef str key
        while True:
            length = zigzag_decode_long(buffer)
            if length == 0:
                return out
            while length:
                value = self.item_type.binary_buffer_decode(buffer)
                out.append(value)
                length -= 1

    cdef _json_format(self, value):
        return [self.item_type.json_format(item) for item in value]

    cdef _json_decode(self, value):
        return [self.item_type.json_decode(item) for item in value]

    cdef int _get_value_fitness(self, value) except -1:
        cdef int level = FIT_OK
        if isinstance(value, (list, tuple, numpy.ndarray)):
            level = FIT_EXACT
        elif isinstance(value, dict):
            level = FIT_POOR
        elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
            return FIT_POOR
        elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
            return FIT_NONE
        try:
            it = iter(value)
        except TypeError:
            return FIT_NONE
        for item in it:
            level = min(self.item_type.get_value_fitness(item), level)
            if level <= FIT_NONE:
                break
        return level

    cdef list _make_converted_list(self, value):
        cdef AvroType item_type = self.item_type
        cdef list out = []
        for item in value:
            out.append(item_type.convert_value(item))
        return out

    cpdef object convert_value(self, object value, check_value=True):
        cdef int item_fitness
        cdef AvroType item_type = self.item_type

        if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
            return []

        it = iter(value)
        for item in it:
            item_fitness = item_type.get_value_fitness(item)
            if item_fitness == FIT_NONE:
                raise ValueError(f"'{item}' is not a valid value for array")
            elif item_fitness < FIT_EXACT:
                return self._make_converted_list(iter(value))
        return value

    cdef _CanonicalForm canonical_form(self, set created):
        return dict_to_canonical({
            'type': 'array',
            'items': self.item_type.canonical_form(created)
        })

    cdef AvroType _for_writer(self, AvroType writer):
        cdef ArrayType cloned
        if isinstance(writer, ArrayType):
            promoted_item = self.item_type._for_writer(writer.item_type)
            if promoted_item is not None:
                cloned = self.clone_base()
                cloned.item_type = promoted_item
                return cloned

    cdef bint accepts_missing_value(self):
        if self.options.missing_values_can_be_empty_container:
            return True


import cavro
import pytest


def test_promote_bytes_to_str():
    writer_schema = cavro.Schema('"string"')
    reader_schema = cavro.Schema('"bytes"')
    resolved = reader_schema.reader_for_writer(writer_schema)
    encoded = writer_schema.binary_encode("hello")
    assert resolved.binary_decode(encoded) == b"hello"


@pytest.mark.parametrize(
    "writer_schema,reader_schema,value,expected",
    [
        ('"int"', '"long"', 1, 1),
        ('"int"', '"float"', 2, 2.),
        ('"int"', '"double"', 3, 3.),
        ('"long"', '"float"', 4, 4.),
        ('"long"', '"double"', 5, 5.),
        ('"float"', '"double"', 6., 6.),
        ({"type": "array", "items": "int"}, {"type": "array", "items": "float"}, [1, 2, 3], [1., 2., 3.]),
        ({"type": "map", "values": "int"}, {"type": "map", "values": "float"}, {'a': 1}, {'a': 1.}),

    ]
)
def test_simple_promotions(writer_schema, reader_schema, value, expected):
    writer_schema = cavro.Schema(writer_schema)
    reader_schema = cavro.Schema(reader_schema)
    resolved = reader_schema.reader_for_writer(writer_schema)
    encoded = writer_schema.binary_encode(value)
    assert resolved.binary_decode(encoded) == expected


@pytest.mark.parametrize(
    "writer_schema,reader_schema",
    [
        ('"long"', '"int"'),
        ('"float"', '"int"'),
        ('"double"', '"int"'),
        ('"float"', '"long"'),
        ('"double"', '"long"'),
        ('"double"', '"float"'),
    ]
)
def test_invalid_promotions(writer_schema, reader_schema):
    writer_schema = cavro.Schema(writer_schema)
    reader_schema = cavro.Schema(reader_schema)
    with pytest.raises(cavro.CannotPromoteError) as exc_info:
        reader_schema.reader_for_writer(writer_schema)
    assert exc_info.value.reader_type is reader_schema.type
    assert exc_info.value.writer_type is writer_schema.type


def test_record_resolution_dict():
    reader = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "long"},
            {"name": "b", "type": "bytes"},
            {"name": "c", "type": "string", "default": "world"},
        ]
    }, record_decodes_to_dict=True)
    writer = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "b", "type": "string"},
        ]
    })
    resolved = reader.reader_for_writer(writer)

    encoded = writer.binary_encode({"a": 1, "b": "hello"})
    decoded = resolved.binary_decode(encoded)
    assert decoded == {"a": 1, "b": b"hello", "c": "world"}


def test_record_resolution_record():
    reader = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "long"},
            {"name": "b", "type": "bytes"},
            {"name": "c", "type": "string", "default": "world"},
        ]
    }, record_decodes_to_dict=False)
    writer = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "b", "type": "string"},
        ]
    })
    resolved = reader.reader_for_writer(writer)

    encoded = writer.binary_encode({"a": 1, "b": "hello"})
    decoded = resolved.binary_decode(encoded)
    assert decoded == reader.named_types['test'].record(a=1, b=b'hello', c='world')


def test_record_resolution_alias():
    reader = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "long"},
            {"name": "b", "type": "bytes", "aliases": ["c"]},
        ]
    }, record_decodes_to_dict=True)
    writer = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "c", "type": "string"},
        ]
    })
    resolved = reader.reader_for_writer(writer)

    encoded = writer.binary_encode({"a": 1, "c": "hello"})
    decoded = resolved.binary_decode(encoded)
    assert decoded == {"a": 1, "b": b"hello"}


def test_field_resolution():
    writer = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "b", "type": "string"},
            {"name": "c", "type": "long"},
        ]
    })
    reader = cavro.Schema({
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "c", "type": "long"},
            {"name": "a", "type": "int"},
        ]
    })
    resolved = reader.reader_for_writer(writer)
    rec = {'a': 1, 'b': 'hello', 'c': 2}
    encoded = writer.binary_encode(rec)
    decoded = resolved.binary_decode(encoded)
    assert decoded._asdict() == {'a': 1, 'c': 2}


import json
import hashlib

cdef str resolve_namespaced_name(str namespace, str name):
    if '.' in name or namespace is None:
        return name
    return f'{namespace}.{name}'


class _class_inst_method:
    def __init__(self, func):
        self.func = func

    def __get__(self, inst, cls):
        return partial(self.func, inst, cls)


cdef class Schema:

    """
    The main interface for `cavro`.

    This class represents an avro schema, and is able to encode and decode values appropriately.

    Arguments:
     * `source`:
        The source of the schema. This can either be a string that holds the JSON-encoded schema definition, or a python object that represents the schema (e.g. the result of `json.loads`).
     * `options`:
        An instance of `Options` that controls how the schema is interpreted. Defaults to `DEFAULT_OPTIONS`.
     * `named_types`:
        An optional dictionary that will be updated to contain any named types that are encountered while parsing the schema.
     * `parse_json`:
        If `False` then the `source` argument will never be parsed as json, even if it's a string value. Defaults to `True`.
     * `**extra_options`:
        Any extra options that should be applied to the schema. These will override any options that are set in the `options` argument.
        Key-values here must match the attributes of `cavro.Options`.
    """


    cdef readonly dict named_types
    cdef readonly object source
    cdef readonly Options options
    cdef readonly AvroType type
    cdef str _canonical_form

    cdef readonly dict logical_types

    def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):
        if isinstance(source, (str, bytes)) and parse_json:
            source = json.loads(source)
        self.source = source
        if extra_options:
            options = options.replace(**extra_options)
        self.options = options
        self.named_types = {} if named_types is None else named_types
        self.logical_types = self._make_logical_types(options)
        self.type = AvroType.for_schema(self) if _type is None else _type

    @_class_inst_method
    def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
        """
        Used in certain situations to create a `Schema` object from a raw `AvroType` instance.
        You probably don't need to use this.
        """
        if inst is not None:
            if options is None:
                options = inst.options
        if options is None:
            options = DEFAULT_OPTIONS
        source = avro_type.get_schema(set())
        cdef Schema new_inst = cls(source, options, parse_json=False, _type=avro_type)
        return new_inst

    def _make_logical_types(self, options):
        logical_by_name = {}
        self.logical_types = logical_by_name
        for logical_type in options.logical_types:
            type_name = logical_type.logical_name
            dest = logical_by_name.setdefault(type_name, [])
            dest.append(logical_type)
        return logical_by_name
    
    cdef void register_type(self, str namespace, str name, AvroType avro_type):
        resolved = resolve_namespaced_name(namespace, name)
        if self.options.named_type_names_must_be_unique and resolved in self.named_types:
            existing = self.named_types[resolved]
            raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')
        self.named_types[resolved] = avro_type

    property canonical_form:
        """
        Returns the canonical form of the schema as a string
        """
        def __get__(self):
            # We are effectively immutable, so caching canonical form is safe here.
            if self._canonical_form is None:
                self._canonical_form = self.type.canonical_form(set())
            return self._canonical_form

    def __hash__(self):
        return hash(self.canonical_form)

    property schema:
        """
        Return an object representing the schema definition.
        Note: This will not always be identical to the `source` used to construct this schema object, as it is reconstructed from the types on-demand.
        """
        def __get__(self):
            return self.type.get_schema(set())

    property schema_str:
        """
        `Schema.schema`, but json encoded
        """
        def __get__(self):
            return json.dumps(self.schema, indent=2)

    def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:
        """
        Return the deterministic fingerprint of the schema, using the given hash method.
        
        `**kwargs` are passed to the relevant `hashlib.new()` call.
        
        Return type is controlled by the `fingerprint_returns_digest` option.
        """
        if method == 'rabin':
            hasher = Rabin()
        else:
            try:
                hasher = hashlib.new(method, **kwargs)
            except ValueError:
                raise InvalidHasher(f'Unknown hash method: {method!r}')
        hasher.update(self.canonical_form.encode('utf-8'))
        if self.options.fingerprint_returns_digest:
            return hasher.digest()
        return hasher

    cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):
        """
        Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
        """
        cdef str resolved = resolve_namespaced_name(namespace, name)
        cdef AvroType found = self.named_types.get(resolved)
        if found is not None:
            return found
        found = self.options.externally_defined_types.get(resolved)
        if found is not None:
            self.register_type(namespace, name, found)
            return found
        found = self.named_types.get(name)
        if found is not None:
            return found
        found = self.options.externally_defined_types.get(name)
        if found is not None:
            return found
        if _raise:
            raise UnknownType(f'Unknown type: {resolved!r}')

    def can_encode(self, value: object) -> bool:
        """
        Check if `value` can be encoded using this schema
        """
        fitness = self.type.get_value_fitness(value)
        return fitness > FIT_NONE

    def binary_encode(self, value: object) -> bytes:
        """
        Encode `value` using this schema and return the avro bytes representing it.
        """
        cdef MemoryWriter buffer = MemoryWriter()
        self.type.binary_buffer_encode(buffer, value)
        return buffer.bytes()

    def binary_decode(self, bytes value: bytes) -> object:
        """
        Decode `value` using this schema and return the decoded value.
        """
        cdef MemoryReader buffer = MemoryReader(value)
        return self.type.binary_buffer_decode(buffer)

    cpdef binary_read(self, _Reader reader: _Reader):
        """
        Read a value from `reader` using this schema and return the decoded value.
        `reader` may be a `MemoryReader` or `FileReader` instance.
        """
        return self.type.binary_buffer_decode(reader)

    cpdef binary_write(self, _Writer writer: _Writer, value: object):
        """
        Write `value` to `writer` using this schema.
        `writer` may be a `MemoryWriter` or `FileWriter` instance.
        """
        self.type.binary_buffer_encode(writer, value)

    def json_encode(self, value, serialize=True, **kwargs):
        """
        Encode `value` using this schema and return the avro json representing it.
        """
        data = self.type.json_format(value)
        if serialize:
            return json.dumps(data, **kwargs)
        return data

    def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):
        """
        Decode `value` in JSON form using this schema and return the decoded value.

        If `deserialize` is True, then value must be a `str` containing the serialized JSON value.
        If `deserialize` is False, then value must be a python object representing the JSON value.
        """
        if deserialize:
            value = json.loads(value, **kwargs)
        return self.type.json_decode(value)

    cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):
        """
        Return a schema that is the result of promoting this schema to the writer schema.
        
        The returned schema may only be used for reading, and should return values that match the reader schema.
        """
        new_type = self.type.for_writer(writer_schema.type)
        return ResolvedSchema(new_type, self.options)



cdef class ResolvedSchema(Schema):

    """
    A variant of a schema that is the result of schema promotion.
    """

    def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):
        named_types = {}
        for sub_type in resolved_type.walk_types(set()):
            if isinstance(sub_type, _NamedType):
                named_types[sub_type.type] = sub_type
            
        self.named_types = named_types
        self.source = None
        self.options = options
        self.type = resolved_type
        self.logical_types = self._make_logical_types(options)


cdef class ValueAdapter:

    """
    Abstract base class for any helper that affects how values are transformed prior to avro encoding/decoding.
    """


    cdef encode_value(self, value):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement encode_value")

    cdef decode_value(self, value):
        raise NotImplementedError(
            f"{type(self).__name__} does not implement decode_value")


cdef int twos_complement(int value, int bits):
    cdef int mask = (1 << bits) - 1
    value = (value ^ mask) + 1
    return value & mask


cdef class LogicalType(ValueAdapter):

    """
    Semi-abstract class for all logical types.
    
    Subclasses must be implemented as cython classes.
    """

    logical_name = NotImplemented
    underlying_types = NotImplemented

    @_class_inst_method
    def for_type(inst, cls, underlying: AvroType):
        cdef object ob = inst
        if inst is None:
            ob = cls

        for underlying_type in ob.underlying_types:
            if isinstance(underlying, underlying_type):
                inst = ob._for_type(underlying)
                if inst is not None:
                    return inst


cdef class CustomLogicalType(LogicalType):

    """
    Logical type that allows custom encoding/decoding functions to be provided.
    
    To use a custom logical type, subclass this class, and implement the:
     * `encode_value(value: object) -> object` - Transforms a provided value into something that can be encoded by the underlying type.
     * `decode_value(value: object) -> object` - Transforms a value decoded by the underlying type into a value to return to the user.
    methods.

    Also implement two class attributes, and a classmethod:
     * `logical_name` - The name to look for in the schema
     * `underlying_types` A tuple of `AvtoType` classes that this logical type can be applied to.
     * `_for_type(cls, underlying: AvroType) -> Cls` 
        A classmethod that returns an instance of the class, optionally customized with information from the underlying type,
        or None if the logical type is not applicable to the underlying type.
    """

    cdef encode_value(self, value):
        return self.custom_encode_value(value)

    cdef decode_value(self, value):
        return self.custom_decode_value(value)


cdef class DecimalType(LogicalType):
    """Logical type for decimal values."""

    logical_name = 'decimal'
    underlying_types = (BytesType, FixedType)

    cdef readonly int precision
    cdef readonly int scale
    cdef readonly object scale_val
    cdef readonly object context
    cdef readonly object size
    cdef str underlying_name

    cdef readonly bint check_exp_overflow

    def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):
        self.precision = precision
        self.scale = scale
        self.scale_val = decimal.Decimal('1').scaleb(-scale)
        self.context = decimal.Context(prec=self.precision, clamp=1)
        self.size = size
        self.check_exp_overflow = check_exp_overflow
        self.underlying_name = underlying_name

    @property
    def type_name(self):
        return f'{self.underlying_name}(decimal)'

    @classmethod
    def _for_type(cls, underlying: AvroType):
        meta = underlying.metadata
        precision = meta.get('precision')
        if precision is None:
            return
        if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
            if underlying.options.raise_on_invalid_logical:
                raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
            return
        scale = meta.get('scale', 0)
        if not isinstance(scale, int) or scale < 0:
            if underlying.options.raise_on_invalid_logical:
                raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
            return
        
        if scale > precision:
            if underlying.options.raise_on_invalid_logical:
                raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
            return
        
        size = None
        if isinstance(underlying, FixedType):
            size = underlying.size
            max_precision = math.floor(math.log10(2) * (8 * size - 1))
            if precision > max_precision: # Numbers may not fit into the fixed
                if underlying.options.raise_on_invalid_logical:
                    raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
                return None

        return cls(precision, scale, size, underlying.options.decimal_check_exp_overflow, underlying.type_name)

    cdef encode_value(self, value):
        if isinstance(value, bytes):
            return value
        if not isinstance(value, decimal.Decimal):
            value = decimal.Decimal(value)
        
        if self.check_exp_overflow:
            _, _, orig_exp = value.as_tuple()
            if orig_exp < -self.scale:
                raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")
        
        quant = value.quantize(self.scale_val, context=self.context)
        
        sign, digits, exp = quant.as_tuple()
        assert exp == -self.scale

        coeff = 0
        for d in digits:
            coeff = coeff * 10 + d
        if sign:
            coeff = -coeff
        
        if self.size is not None:
            return coeff.to_bytes(self.size, byteorder='big', signed=True)
        else:
            n_bytes = (coeff.bit_length() + 8) // 8
            return coeff.to_bytes(n_bytes, byteorder='big', signed=True)

    cdef decode_value(self, value):
        coeff = int.from_bytes(value, byteorder='big', signed=True)
        decimal_val = self.context.create_decimal(coeff)
        return decimal_val.scaleb(-self.scale, context=self.context)

      

cdef class UUIDBase(LogicalType):
    """Logical type for UUID values"""
    logical_name = 'uuid'
    cdef bint return_uuid_object

    def __init__(self, return_uuid_object):
        self.return_uuid_object = return_uuid_object

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls(underlying.options.return_uuid_object)


cdef class UUIDStringType(UUIDBase):
    underlying_types = (StringType, )

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls(underlying.options.return_uuid_object)

    @property
    def type_name(self):
        return 'string(uuid)'

    cdef encode_value(self, value):
        # apache avro lib behaviour is weird here, 
        # value should be validated, but unmodified value should
        # be encoded (i.e. the presence of '-' or not should be retained)
        if isinstance(value, uuid.UUID):
            return str(value)
        try:
            _ = uuid.UUID(hex=value)
        except TypeError as e:
            raise InvalidValue(value, self) from e
        return value

    cdef decode_value(self, value):
        if self.return_uuid_object:
            return uuid.UUID(value)
        return value
        

cdef class UUIDFixedType(UUIDBase):
    underlying_types = (FixedType, )

    @classmethod
    def _for_type(cls, underlying: AvroType):
        if underlying.size != 16:
            return None
        return cls(underlying.options.return_uuid_object)

    @property
    def type_name(self):
        return 'fixed(uuid)'

    cdef encode_value(self, value):
        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        return value.bytes

    cdef decode_value(self, value):
        if self.return_uuid_object:
            return uuid.UUID(bytes=value)
        return value


EPOCH_DATE = datetime.date(1970, 1, 1)
EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)

cdef class Date(LogicalType):
    """Logical type for Date values"""
    logical_name = 'date'
    underlying_types = (IntType, )

    cdef readonly bint accepts_string

    def __init__(self, accepts_string):
        self.accepts_string = accepts_string

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls(underlying.options.date_type_accepts_string)

    @property
    def type_name(self):
        return 'int(date)'

    cdef encode_value(self, value):
        if self.accepts_string and isinstance(value, str):
            value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
        if isinstance(value, int) and value >= 0 and value <= 2**16:
            return value
        if isinstance(value, datetime.datetime):
            value = value.date()
        if isinstance(value, datetime.date):
            return (value - EPOCH_DATE).days
        raise ValueError(f"Expected datetime.date, got {type(value)}")

    cdef decode_value(self, value):
        return EPOCH_DATE + datetime.timedelta(days=value)


def _time_to_micros(value):
    return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))


def _micros_to_time(value):
    return datetime.time(
        hour=value // 3_600_000_000,
        minute=(value // 60_000_000) % 60,
        second=(value // 1_000_000) % 60,
        microsecond=value % 100_0000
    )


MAX_TIME = _time_to_micros(datetime.time.max)


cdef object dt_to_timedelta(value):
    if value.tzinfo is None:
        value = value.astimezone(datetime.timezone.utc)
    return value - EPOCH_DT


cdef class TimeMillis(LogicalType):
    """Logical type for time-millis values"""
    logical_name = 'time-millis'
    underlying_types = (IntType, )

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls()

    @property
    def type_name(self):
        return 'int(time-millis)'

    cdef encode_value(self, value):
        if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
            return value
        if isinstance(value, datetime.datetime):
            value = value.time()
        return _time_to_micros(value) // 1_000

    cdef decode_value(self, value):
        return _micros_to_time(value * 1_000)


cdef class TimeMicros(LogicalType):
    """Logical type for time-micros values"""
    logical_name = 'time-micros'
    underlying_types = (LongType, )

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls()

    @property
    def type_name(self):
        return 'long(time-micros)'

    cdef encode_value(self, value):
        if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
            return value
        if isinstance(value, datetime.datetime):
            value = value.time()
        return _time_to_micros(value)

    cdef decode_value(self, value):
        return _micros_to_time(value)


cdef class TimestampMillis(LogicalType):
    """Logical type for timestamp-micros values"""
    logical_name = 'timestamp-millis'
    underlying_types = (LongType, )

    cdef readonly bint alternate_timestamp_encoding

    def __init__(self, alternate_timestamp_encoding):
        self.alternate_timestamp_encoding = alternate_timestamp_encoding

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls(
            underlying.options.alternate_timestamp_millis_encoding,
        )

    @property
    def type_name(self):
        return 'long(timestamp-millis)'

    cdef encode_value(self, value):
        if isinstance(value, int):
            return value
        if not isinstance(value, datetime.datetime):
            raise ValueError(f"Expected datetime.datetime, got {type(value)}")
        if self.alternate_timestamp_encoding:
            delta = dt_to_timedelta(value)
            return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000
        else:
            return int(value.timestamp() * 1_000)

    cdef decode_value(self, value):
        return EPOCH_DT + datetime.timedelta(microseconds=value * 1_000)


cdef class TimestampMicros(LogicalType):
    """Logical type for timestamp-micros values"""
    logical_name = 'timestamp-micros'
    underlying_types = (LongType, )

    @classmethod
    def _for_type(cls, underlying: AvroType):
        return cls()

    @property
    def type_name(self):
        return 'long(timestamp-micros)'

    cdef encode_value(self, value):
        if isinstance(value, int):
            return value
        if not isinstance(value, datetime.datetime):
            raise ValueError(f"Expected datetime.datetime, got {type(value)}")
        delta = dt_to_timedelta(value)
        return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000

    cdef decode_value(self, value):
        return EPOCH_DT + datetime.timedelta(microseconds=value)




import glob
import os
import hashlib
from os import path
import re
import subprocess
import sys

from setuptools import setup
from distutils.command.build_clib import build_clib
from setuptools.extension import Extension


PROJECT_ROOT = path.dirname(path.abspath(__file__))


with open(path.join(PROJECT_ROOT, "README.md")) as fh:
    LONG_DESCRIPTION = fh.read()


sources = ['cavro.c']
ext_args = {
    'extra_compile_args': ['-g', '-O2'], 
    'extra_link_args': ['-g'],
}

try:
    from Cython.Build import cythonize
except ImportError:
    ext = [Extension('cavro', sources=['cavro.c'], **ext_args)]
else:
    ext = cythonize(
        Extension('cavro', sources=['cavro.pyx'], **ext_args),
        compiler_directives={"language_level": 3},
    )


setup(
    name='cavro',
    ext_modules = ext,
    cmdclass = {'build_clib': build_clib},
    version='1.0.0',
    description="A Cython based Avro library",
    long_description=LONG_DESCRIPTION,
    author="Stephen Stagg",
    author_email="stestagg@gmail.com",
    python_requires=">=3.8.0",
    url="https://cavro.io/",
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)




"""
`cavro` is a library for encoding and decoding data in the Apache Avro format.

It's written in Cython, with a focus on performance, correctness and ease-of-use.
"""

cimport cython
import warnings
from cpython.object cimport Py_SIZE
from libc.stdint cimport *
from libc.string cimport memcmp
import inspect
import dataclasses

from collections.abc import Sequence

from cython.dataclasses cimport dataclass
import datetime
import re
import decimal
import enum
from cpython cimport array
from cython cimport bint
from functools import partial
import uuid
import math
from types import MappingProxyType
from typing import Union
from cpython.dict cimport PyDictProxy_New

from cython.view cimport array as cvarray

from libc.string cimport memcpy

ctypedef bint bool

if cython.sizeof(Py_ssize_t) == 8:
    SSIZE_TYPECODE = b'q'
elif cython.sizeof(Py_ssize_t) == 4:
    SSIZE_TYPECODE = b"l"
else:
    SSIZE_TYPECODE = "UNKNOWN"

__version__ = "1.0.0"

__tests = {}
def _tests(fn):
    def add(test, *args):
        name = test.__name__
        if args:
            name += '-' + ('-'.join(str(a) for a in args))
        __tests[name] = (test, ) + args
    fn(add)
    return _tests

__perf_tests = {}
def _perf(fn):
    def add(test, *args):
        name = test.__name__
        if args:
            name += '-' + ('-'.join(str(a) for a in args))
        __perf_tests[name] = (test, ) + args
    fn(add)
    return _perf

include "src/error.pxi"
include "src/option.pxi"
include "src/io.pxi"
include "src/buffer.pxi"
include "src/zigzag.pxi"
include "src/rabin.pxi"

include "src/tests/test_zigzag.pxi"
include "src/tests/test_buffer.pxi"
include "src/tests/test_perf.pxi"

include "src/array.pxi"
include "src/enum.pxi"
include "src/map.pxi"
include "src/null.pxi"
include "src/numeric_types.pxi"
include "src/record.pxi"
include "src/string_types.pxi"
include "src/union.pxi"

include "src/type.pxi"
include "src/schema.pxi"

include "src/logical.pxi"

include "src/codec.pxi"
include "src/container.pxi"

include "src/promotions.pxi"

